# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from importlib.metadata import PackageNotFoundError, version
    __version__ = version('bazis-permit')
except PackageNotFoundError:
    __version__ = 'dev'


"""
TODO: there is a problem now:
if the add action is available for an object, it can be added with a reference to
an external object that the user does not have access to.
this happens because the object does not have filled selectors at the addition stage.

we need a solution that will allow checking the availability of related objects
at the stage of adding or editing the main object.

that is, we need a solution that will make such a permission work:
'entity.extended_entity.item.add.author_parent', # only those who are the author of the
parent record can add entries

"""



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import quote, urlencode

import pytest
from bazis_test_utils.utils import get_api_client
from translated_fields import to_attribute

from bazis.contrib.permit.models import GroupPermission, Permission, Role
from bazis.contrib.users import get_user_model

from tests import factories


User = get_user_model()


GROUPS = {
    'parent_entity': [
        'entity.parent_entity.item.add.all',  # everyone has the ability to add records
        'entity.parent_entity.item.view.author',  # everyone has the ability to view any records
        'entity.parent_entity.item.change.author',  # only the author can edit the record
    ],
    'parent_entity_view_all': [
        'entity.parent_entity.item.add.all',  # everyone has the ability to add records
        'entity.parent_entity.item.view.all',  # everyone has the ability to view any records
        'entity.parent_entity.item.change.author',  # only the author can edit the record
    ],
    'child_entity': [
        'entity.child_entity.item.add.all',  # everyone has the ability to add records
        'entity.child_entity.item.view.all',  # everyone has the ability to view any records
        'entity.child_entity.item.change.author',  # only the author can edit the record
        'entity.child_entity.item.change.author_parent',  # the parent author can edit the record
        'entity.child_entity.item.delete.author',  # only the author can delete the record
    ],
    'dependent_entity': [
        'entity.dependent_entity.item.add.all',  # everyone has the ability to add records
        'entity.dependent_entity.item.view.all',  # everyone has the ability to view any records
        'entity.dependent_entity.item.change.author',  # only the author can edit the record
        'entity.dependent_entity.item.delete.author',  # only the author can delete the record
    ],
    'extended_entity': [
        'entity.extended_entity.item.add.all',  # everyone has the ability to add records
        'entity.extended_entity.item.view.author',  # only the author can view the record
        'entity.extended_entity.item.view.author_parent',  # the parent author can edit the record
        'entity.extended_entity.item.change.author',  # only the author can edit the record
        'entity.extended_entity.item.delete.author',  # only the author can delete the record
    ],
    'complex_parent_entity': [
        'entity.parent_entity.item.add.all',
        f'entity.parent_entity.item.check.author=__selector__&name={quote("Test name")}',
        'entity.parent_entity.item.view.author=__selector__',
        'entity.parent_entity.item.change.author=__selector__&is_active=false',
        'entity.parent_entity.field.view.author=__selector__&is_active=true.description.enable',
        'entity.parent_entity.field.view.author=__selector__&is_active=true.__all__.disable',
        'entity.parent_entity.field.view.author=__selector__&is_active=false.description.disable',
    ],
    'child_entity_fields': [
        'entity.child_entity.item.add.all',
        'entity.child_entity.field.add.all.child_name.filter:^[A-Z]+$',
        'entity.parent_entity.field.add.all.child_entities.filter:child_is_active=true',
        'entity.child_entity.field.change.all.child_name.filter:^[A-Z]+$',
        'entity.parent_entity.field.change.all.child_entities.filter:child_is_active=true',
        'entity.dependent_entity.field.add.all.parent_entity.filter:is_active=true',
        'entity.dependent_entity.field.change.all.parent_entity.filter:is_active=true',
    ],
}


@pytest.fixture(scope='function')
def groups():
    # create roles and groups
    groups = {}
    for group_name, permissions in GROUPS.items():
        group = GroupPermission.objects.create(
            slug=group_name,
            **{to_attribute('name'): group_name}
        )
        for it in permissions:
            group.permissions.add(Permission.objects.get_or_create(slug=it)[0])
        groups[group_name] = group
    return groups


def role_create(groups, name: str, groups_names: list[str]):
    """
    Helper function to create a role and assign it to specified groups of
    permissions.
    """
    # create roles and groups
    role = Role.objects.create(slug=name, **{to_attribute('name'): name})
    for group_name in groups_names:
        role.groups_permission.add(groups[group_name])
    return role


@pytest.mark.django_db(transaction=True)
def test_transfer_trigger():
    """
    Test the transfer trigger functionality by creating a parent entity and a child
    entity, linking them, and verifying the parent author relationship.
    """
    user_1 = User.objects.create_user('user1', email='user1@site.com', password='weak_password_2')
    parent_entity = factories.ParentEntityFactory.create(author=user_1)
    child_entity = factories.ChildEntityFactory.create()
    parent_entity.child_entities.add(child_entity)
    child_entity.save()
    child_entity.refresh_from_db()
    assert child_entity.author_parent == parent_entity.author


@pytest.mark.django_db(transaction=True)
def test_permit(sample_app, groups):
    """
    Test the permission system by verifying the ability to create, view, edit, and
    delete records based on user roles and permissions.
    """
    # without authorization, you cannot view/patch/create/delete records on protected routes
    parent_entity = factories.ParentEntityFactory.create(child_entities=True)

    response = get_api_client(sample_app).get(f'/api/v1/entity/parent_entity/{parent_entity.id}/schema_retrieve/')
    assert response.status_code == 403

    response = get_api_client(sample_app).get(f'/api/v1/entity/parent_entity/{parent_entity.id}/')
    assert response.status_code == 403

    response = get_api_client(sample_app).patch(
        f'/api/v1/entity/parent_entity/{parent_entity.id}/',
        json_data={
            'data': {
                'id': str(parent_entity.id),
                'type': 'entity.parent_entity',
                'bs:action': 'change',
                'attributes': {
                    'name': 'New test name',
                },
            },
        },
    )
    assert response.status_code == 403

    response = get_api_client(sample_app).post(
        '/api/v1/entity/parent_entity/',
        json_data={
            'data': {
                'type': 'entity.parent_entity',
                'bs:action': 'add',
                'attributes': {
                    'name': 'Test name',
                    'description': 'Test description',
                    'is_active': True,
                    'price': '100.49',
                    'dt_approved': '2024-01-12T16:54:12Z',
                },
            },
        },
    )
    assert response.status_code == 403

    response = get_api_client(sample_app).delete(
        f'/api/v1/entity/parent_entity/{str(parent_entity.id)}/'
    )
    assert response.status_code == 403

    user_1 = User.objects.create_user('user1', email='user1@site.com', password='weak_password_1')
    user_2 = User.objects.create_user('user2', email='user2@site.com', password='weak_password_2')
    user_3 = User.objects.create_user('user3', email='user3@site.com', password='weak_password_3')

    # user_1 is assigned the role role_user1, which has access to the parent_entity
    # and child_entity groups
    user_1.roles.add(
        role_create(
            groups,
            'role_user1',
            [
                'parent_entity',
                'child_entity',
                'dependent_entity',
                'extended_entity',
            ],
        )
    )
    user_2.roles.add(
        role_create(
            groups,
            'role_user2',
            [
                'parent_entity_view_all',
                'child_entity',
                'dependent_entity',
                'extended_entity',
            ],
        )
    )
    user_3.roles.add(role_create(groups,'role_user3', ['child_entity']))

    # now user_1 can create records in parent_entity
    response = get_api_client(sample_app, user_1.jwt_build()).post(
        '/api/v1/entity/parent_entity/',
        json_data={
            'data': {
                'type': 'entity.parent_entity',
                'bs:action': 'add',
                'attributes': {
                    'name': 'Test name',
                    'description': 'Test description',
                    'is_active': True,
                    'price': '100.49',
                    'dt_approved': '2024-01-12T16:54:12Z',
                },
            },
        },
    )
    assert response.status_code == 201

    data = response.json()

    it = data['data']
    assert it['type'] == 'entity.parent_entity'
    assert it['bs:action'] == 'view'

    parent_entity_id = str(it['id'])

    response = get_api_client(sample_app).get(
        '/api/v1/entity/parent_entity/',
    )
    assert response.status_code == 403


    response = get_api_client(sample_app, user_1.jwt_build()).get(
        '/api/v1/entity/parent_entity/',
    )
    assert response.status_code == 200
    assert len(response.json()['data']) == 1


    # user user_2 can view records in parent_entity
    response = get_api_client(sample_app, user_2.jwt_build()).get(
        f'/api/v1/entity/parent_entity/{parent_entity_id}/'
    )
    assert response.status_code == 200

    # user user_1 can edit the record in parent_entity
    response = get_api_client(sample_app, user_1.jwt_build()).patch(
        f'/api/v1/entity/parent_entity/{parent_entity_id}/',
        json_data={
            'data': {
                'id': parent_entity_id,
                'type': 'entity.parent_entity',
                'bs:action': 'change',
                'attributes': {
                    'name': 'Test name',
                    'description': 'Test description',
                    'is_active': True,
                    'price': '100.49',
                    'dt_approved': '2024-01-12T16:54:12Z',
                },
            },
        },
    )
    assert response.status_code == 200

    # user user_2 does not have access to edit parent_entity
    response = get_api_client(sample_app, user_2.jwt_build()).patch(
        f'/api/v1/entity/parent_entity/{parent_entity_id}/',
        json_data={
            'data': {
                'id': parent_entity_id,
                'type': 'entity.parent_entity',
                'bs:action': 'change',
                'attributes': {
                    'name': 'Test name',
                    'description': 'Test description',
                    'is_active': True,
                    'price': '100.49',
                    'dt_approved': '2024-01-12T16:54:12Z',
                },
            },
        },
    )
    assert response.status_code == 403

    # user user_1 cannot delete the record in parent_entity
    response = get_api_client(sample_app, user_1.jwt_build()).delete(
        f'/api/v1/entity/parent_entity/{parent_entity_id}/'
    )
    assert response.status_code == 403

    # user user_3 creates a child record, referencing parent_entity_id
    response = get_api_client(sample_app, user_3.jwt_build()).post(
        '/api/v1/entity/child_entity/',
        json_data={
            'data': {
                'type': 'entity.child_entity',
                'bs:action': 'add',
                'attributes': {
                    'child_name': 'Child test name 2',
                    'child_description': 'Child test description 2',
                    'child_is_active': False,
                    'child_price': '25.19',
                    'child_dt_approved': '2024-01-13T16:54:12Z',
                },
                'relationships': {
                    'parent_entities': {
                        'data': [
                            {
                                'id': parent_entity_id,
                                'type': 'entity.parent_entity',
                            }
                        ],
                    },
                },
            },
        },
    )
    assert response.status_code == 201
    child_entity_id = str(response.json()['data']['id'])

    # since the relationship between child_entity and parent_entity is m2m,
    # child_entity needs to be updated to refresh author_parent
    response = get_api_client(sample_app, user_3.jwt_build()).patch(
        f'/api/v1/entity/child_entity/{child_entity_id}/',
        json_data={
            'data': {
                'id': child_entity_id,
                'type': 'entity.child_entity',
                'bs:action': 'change',
                'attributes': {},
            },
        },
    )
    assert response.status_code == 200

    data = response.json()
    relationships = data['data']['relationships']

    # check that the parent author matches the parent author of the child record
    assert relationships['author_parent'] == {
        'data': {
            'id': str(user_1.pk),
            'type': 'users.user',
        },
    }

    # user user_2 cannot edit child_entity
    response = get_api_client(sample_app, user_2.jwt_build()).patch(
        f'/api/v1/entity/child_entity/{child_entity_id}/',
        json_data={
            'data': {
                'id': child_entity_id,
                'type': 'entity.child_entity',
                'bs:action': 'change',
                'attributes': {},
            },
        },
    )
    assert response.status_code == 403

    # user user_1 can edit child_entity thanks to 'entity.child_entity.item.change.author_parent'
    response = get_api_client(sample_app, user_1.jwt_build()).patch(
        f'/api/v1/entity/child_entity/{child_entity_id}/',
        json_data={
            'data': {
                'id': child_entity_id,
                'type': 'entity.child_entity',
                'bs:action': 'change',
                'attributes': {},
            },
        },
    )
    assert response.status_code == 200

    # user_1 can add dependent_entity and extended_entity
    response = get_api_client(sample_app, user_1.jwt_build()).post(
        '/api/v1/entity/dependent_entity/',
        json_data={
            'data': {
                'type': 'entity.dependent_entity',
                'bs:action': 'add',
                'attributes': {
                    'dependent_name': 'Dependent test name',
                    'dependent_description': 'Dependent test description',
                    'dependent_is_active': True,
                    'dependent_price': '100.00',
                    'dependent_dt_approved': '2024-06-28T16:54:12Z',
                },
                'relationships': {
                    'parent_entity': {
                        'data': {
                            'id': parent_entity_id,
                            'type': 'entity.parent_entity',
                        },
                    },
                },
            },
        },
    )

    assert response.status_code == 201

    data = response.json()

    it = data['data']
    assert it['type'] == 'entity.dependent_entity'
    assert it['bs:action'] == 'view'

    assert data['data']['relationships']['author_parent']['data']['id'] == str(user_1.id)
    depended_entity_id = str(it['id'])

    # user_1 can add extended_entity
    response = get_api_client(sample_app, user_1.jwt_build()).post(
        '/api/v1/entity/extended_entity/',
        json_data={
            'data': {
                'type': 'entity.extended_entity',
                'bs:action': 'add',
                'attributes': {
                    'extended_name': 'Extended test name',
                    'extended_description': 'Extended test description',
                    'extended_is_active': True,
                    'extended_price': '100.00',
                    'extended_dt_approved': '2024-06-28T16:54:12Z',
                },
                'relationships': {
                    'parent_entity': {
                        'data': {
                            'id': parent_entity_id,
                            'type': 'entity.parent_entity',
                        },
                    },
                },
            }
        },
    )

    assert response.status_code == 201

    data = response.json()

    it = data['data']
    assert it['type'] == 'entity.extended_entity'
    assert it['bs:action'] == 'view'

    assert data['data']['relationships']['author_parent']['data']['id'] == str(user_1.id)
    extended_entity_id = str(it['id'])

    # user user_2 can view parent_entity and child_entity, dependent_entity,
    # but not extended_entity
    query = urlencode(
        {
            'include': 'extended_entity,dependent_entities,child_entities',
        }
    )

    response = get_api_client(sample_app, user_2.jwt_build()).get(
        f'/api/v1/entity/parent_entity/{parent_entity_id}/?{query}'
    )
    assert response.status_code == 200

    data = response.json()

    it = data['data']
    assert it['type'] == 'entity.parent_entity'
    assert it['bs:action'] == 'view'

    assert parent_entity_id == str(it['id'])

    assert len(data['included']) == 2

    for include in data['included']:
        assert include['type'] in ['entity.child_entity', 'entity.dependent_entity']
        assert include['type'] != 'entity.extended_entity'

    # user user_1 can view parent_entity and child_entity, dependent_entity and extended_entity
    response = get_api_client(sample_app, user_1.jwt_build()).get(
        f'/api/v1/entity/parent_entity/{parent_entity_id}/?{query}'
    )
    assert response.status_code == 200

    data = response.json()

    it = data['data']
    assert it['type'] == 'entity.parent_entity'
    assert it['bs:action'] == 'view'

    assert parent_entity_id == str(it['id'])

    assert len(data['included']) == 3
    for include in data['included']:
        assert include['type'] in [
            'entity.child_entity',
            'entity.dependent_entity',
            'entity.extended_entity',
        ]

    # user user_2 cannot delete the record with extended_entity and dependent_entity
    response = get_api_client(sample_app, user_2.jwt_build()).delete(
        f'/api/v1/entity/extended_entity/{extended_entity_id}/'
    )
    assert response.status_code == 403

    response = get_api_client(sample_app, user_2.jwt_build()).delete(
        f'/api/v1/entity/dependent_entity/{depended_entity_id}/'
    )
    assert response.status_code == 403

    # user user_1 can delete the record with extended_entity and dependent_entity
    response = get_api_client(sample_app, user_1.jwt_build()).delete(
        f'/api/v1/entity/extended_entity/{extended_entity_id}/'
    )
    assert response.status_code == 204

    response = get_api_client(sample_app, user_1.jwt_build()).delete(
        f'/api/v1/entity/dependent_entity/{depended_entity_id}/'
    )
    assert response.status_code == 204


@pytest.mark.django_db(transaction=True)
def test_permit_complex(sample_app, groups):
    """
    Test the permission system by verifying the ability to create, view, edit, and
    delete records based on user roles and permissions.
    """
    user_1 = User.objects.create_user('user1', email='user1@site.com', password='weak_password_1')

    # user_1 is assigned the role role_user1, which has access to the parent_entity
    # and child_entity groups
    user_1.roles.add(
        role_create(
            groups,
            'role_user1',
            [
                'complex_parent_entity',
            ],
        )
    )

    response = get_api_client(sample_app, user_1.jwt_build()).post(
        '/api/v1/entity/parent_entity/',
        json_data={
            'data': {
                'type': 'entity.parent_entity',
                'bs:action': 'add',
                'attributes': {
                    'name': 'Test name 1',
                    'description': 'Test description',
                    'is_active': True,
                    'price': '100.49',
                    'dt_approved': '2024-01-12T16:54:12Z',
                },
            },
        },
    )
    assert response.status_code == 403

    # now user_1 can create records in parent_entity
    for _ in range(2):
        response = get_api_client(sample_app, user_1.jwt_build()).post(
            '/api/v1/entity/parent_entity/',
            json_data={
                'data': {
                    'type': 'entity.parent_entity',
                    'bs:action': 'add',
                    'attributes': {
                        'name': 'Test name',
                        'description': 'Test description',
                        'is_active': False,
                        'price': '100.49',
                        'dt_approved': '2024-01-12T16:54:12Z',
                    },
                },
            },
        )

        assert response.status_code == 201

        data = response.json()

        it = data['data']
        assert it['type'] == 'entity.parent_entity'
        assert it['bs:action'] == 'view'

    item_id = response.json()['data']['id']

    response = get_api_client(sample_app, user_1.jwt_build()).patch(
        f'/api/v1/entity/parent_entity/{item_id}/',
        json_data={
            'data': {
                'id': item_id,
                'type': 'entity.parent_entity',
                'bs:action': 'change',
                'attributes': {
                    'is_active': True,
                },
            },
        },
    )
    assert response.status_code == 200

    # the second attempt to change the record will be unsuccessful because only is_active=false is allowed for editing
    response = get_api_client(sample_app, user_1.jwt_build()).patch(
        f'/api/v1/entity/parent_entity/{item_id}/',
        json_data={
            'data': {
                'id': item_id,
                'type': 'entity.parent_entity',
                'bs:action': 'change',
                'attributes': {
                    'is_active': True,
                },
            },
        },
    )
    assert response.status_code == 403


    response = get_api_client(sample_app, user_1.jwt_build()).get(
        '/api/v1/entity/parent_entity/',
    )
    assert response.status_code == 200

    data_list = response.json()['data']

    assert len(data_list) == 2

    for it in data_list:
        if it['id'] == item_id:
            assert it['attributes'] == {'description': 'Test description'}
        else:
            assert 'description' not in it['attributes']


@pytest.mark.django_db(transaction=True)
def test_filters_fields(sample_app, groups):
    user_1 = User.objects.create_user('user1', email='user1@site.com', password='weak_password_1')

    user_1.roles.add(
        role_create(
            groups,
            'role_user1',
            [
                'child_entity',
                'child_entity_fields',
                'parent_entity',
                'dependent_entity',
            ],
        )
    )

    children_is_active = [True, False, True]

    # create child items with different child_is_active values
    # also pass child_name, which does not match the field filter
    children = []
    for is_active in children_is_active:
        response = get_api_client(sample_app, user_1.jwt_build()).post(
            '/api/v1/entity/child_entity/',
            json_data={
                'data': {
                    'type': 'entity.child_entity',
                    'bs:action': 'add',
                    'attributes': {
                        'child_name': 'Test name 1',
                        'child_description': 'Test description',
                        'child_is_active': is_active,
                        'child_price': '100.49',
                        'child_dt_approved': '2024-01-12T16:54:12Z',
                    },
                },
            },
        )

        data = response.json()

        assert response.status_code == 201

        # make sure that child_name is not set
        assert data['data']['attributes']['child_name'] == ''

        children.append(data['data']['id'])

    # create a parent item, trying to set ALL child items
    response = get_api_client(sample_app, user_1.jwt_build()).post(
        '/api/v1/entity/parent_entity/',
        json_data={
            'data': {
                'type': 'entity.parent_entity',
                'bs:action': 'add',
                'attributes': {
                    'name': 'Test name 1',
                    'description': 'Test description',
                    'is_active': True,
                    'price': '100.49',
                    'dt_approved': '2024-01-12T16:54:12Z',
                },
                'relationships': {
                    'child_entities': {
                        'data': [
                            {
                                'type': 'entity.child_entity',
                                'id': child_id,
                            } for child_id in children
                        ],
                    },
                },
            },
        },
    )

    data = response.json()

    parent_id = data['data']['id']

    child_data_ids = [child['id'] for child in data['data']['relationships']['child_entities']['data']]

    # ONLY child items with child_is_active=True should be set after creating the parent
    for i, child in enumerate(children):
        if children_is_active[i]:
            assert child in child_data_ids
        else:
            assert child not in child_data_ids

    # now set a valid child_name, but deactivate child_is_active
    response = get_api_client(sample_app, user_1.jwt_build()).patch(
        f'/api/v1/entity/child_entity/{children[0]}/',
        json_data={
            'data': {
                'id': children[0],
                'type': 'entity.child_entity',
                'bs:action': 'change',
                'attributes': {
                    'child_name': 'TEST',
                    'child_is_active': False,
                },
            },
        },
    )
    assert response.status_code == 200

    data = response.json()

    # make sure that child_name is set
    assert data['data']['attributes']['child_name'] == 'TEST'

    # explicitly set only 1 child item (the third one)
    response = get_api_client(sample_app, user_1.jwt_build()).patch(
        f'/api/v1/entity/parent_entity/{parent_id}/',
        json_data={
            'data': {
                'id': parent_id,
                'type': 'entity.parent_entity',
                'bs:action': 'change',
                'relationships': {
                    'child_entities': {
                        'data': [
                            {
                                'type': 'entity.child_entity',
                                'id': child_id,
                            } for child_id in children[2:3]
                        ],
                    },
                },
            },
        },
    )

    data = response.json()

    # although only the third one was set, the first one is also available, because at the time of changing
    # the parent, the first item's child_is_active = False, and resetting it
    # is not available to the user, therefore it remains in the set
    assert len(data['data']['relationships']['child_entities']['data']) == 2
    #
    for child in data['data']['relationships']['child_entities']['data']:
        # this item is available because the user cannot change it according to the filter (child_is_active: False)
        if child['id'] == children[0]:
            assert child['id'] in children
        # this item is available because the user can change it according to the filter and it was just set (child_is_active: False)
        if child['id'] == children[2]:
            assert child['id'] in children

    # in this case we can set the parent, because the parent has is_active: True
    response = get_api_client(sample_app, user_1.jwt_build()).post(
        '/api/v1/entity/dependent_entity/',
        json_data={
            'data': {
                'type': 'entity.dependent_entity',
                'bs:action': 'add',
                'attributes': {
                    'dependent_name': 'Dependent test name',
                },
                'relationships': {
                    'parent_entity': {
                        'data':
                            {
                                'type': 'entity.parent_entity',
                                'id': parent_id,
                            }
                        ,
                    },
                },
            },
        },
    )

    data = response.json()

    dependent_entity_id = data['data']['id']

    assert data['data']['relationships']['parent_entity']['data']['id'] == parent_id

    # set the parent to is_active: False
    response = get_api_client(sample_app, user_1.jwt_build()).patch(
        f'/api/v1/entity/parent_entity/{parent_id}/',
        json_data={
            'data': {
                'id': parent_id,
                'type': 'entity.parent_entity',
                'bs:action': 'change',
                'attributes': {
                    'is_active': False,
                },
            },
        },
    )

    parent_wrong = factories.ParentEntityFactory.create(
        name='Parent test name',
    )

    # in this case we cannot change the parent, because the original parent has is_active: False
    response = get_api_client(sample_app, user_1.jwt_build()).patch(
        f'/api/v1/entity/dependent_entity/{dependent_entity_id}/',
        json_data={
            'data': {
                'id': dependent_entity_id,
                'type': 'entity.dependent_entity',
                'bs:action': 'change',
                'relationships': {
                    'parent_entity': {
                        'data':
                            {
                                'type': 'entity.parent_entity',
                                'id': str(parent_wrong.id),
                            }
                        ,
                    },
                },
            },
        },
    )

    data = response.json()

    assert data['data']['relationships']['parent_entity']['data']['id'] == parent_id



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.utils.translation import gettext_lazy as _

from pydantic import Field

from bazis.core.utils.schemas import BazisSettings


class Settings(BazisSettings):
    """
    Represents the settings configuration for the application, extending
    BazisSettings to include a specific permission cache expiry time.
    """

    BAZIS_PERMISSION_CACHE_EXPIRE: int = Field(
        7, title=_('Time to store user permissions, sec'), dynamic=True
    )


settings = Settings()



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 5.0.1 on 2024-10-20 00:05

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('entity', '0003_childentity__author_parent_selectors_and_more'),
    ]

    operations = [
        migrations.RenameField(
            model_name='childentity',
            old_name='_author_parent_selectors',
            new_name='autogen_author_parent_selectors',
        ),
        migrations.RenameField(
            model_name='childentity',
            old_name='_author_selectors',
            new_name='autogen_author_selectors',
        ),
        migrations.RenameField(
            model_name='childentity',
            old_name='_author_updated_selectors',
            new_name='autogen_author_updated_selectors',
        ),
        migrations.RenameField(
            model_name='dependententity',
            old_name='_author_parent_selectors',
            new_name='autogen_author_parent_selectors',
        ),
        migrations.RenameField(
            model_name='dependententity',
            old_name='_author_selectors',
            new_name='autogen_author_selectors',
        ),
        migrations.RenameField(
            model_name='dependententity',
            old_name='_author_updated_selectors',
            new_name='autogen_author_updated_selectors',
        ),
        migrations.RenameField(
            model_name='extendedentity',
            old_name='_author_parent_selectors',
            new_name='autogen_author_parent_selectors',
        ),
        migrations.RenameField(
            model_name='extendedentity',
            old_name='_author_selectors',
            new_name='autogen_author_selectors',
        ),
        migrations.RenameField(
            model_name='extendedentity',
            old_name='_author_updated_selectors',
            new_name='autogen_author_updated_selectors',
        ),
        migrations.RenameField(
            model_name='parententity',
            old_name='_author_selectors',
            new_name='autogen_author_selectors',
        ),
        migrations.RenameField(
            model_name='parententity',
            old_name='_author_updated_selectors',
            new_name='autogen_author_updated_selectors',
        ),
    ]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bazis_test_utils import factories_abstract
from entity.models import ChildEntity, DependentEntity, ExtendedEntity, ParentEntity


class ChildEntityFactory(factories_abstract.ChildEntityFactoryAbstract):
    """
    Factory class for creating instances of ChildEntity, inheriting from
    ChildEntityFactoryAbstract.
    """

    class Meta:
        """
        Meta class specifying the model for ChildEntityFactory as ChildEntity.
        """

        model = ChildEntity


class DependentEntityFactory(factories_abstract.DependentEntityFactoryAbstract):
    """
    Factory class for creating instances of DependentEntity, inheriting from
    DependentEntityFactoryAbstract.
    """

    class Meta:
        """
        Meta class specifying the model for DependentEntityFactory as DependentEntity.
        """

        model = DependentEntity


class ExtendedEntityFactory(factories_abstract.ExtendedEntityFactoryAbstract):
    """
    Factory class for creating instances of ExtendedEntity, inheriting from
    ExtendedEntityFactoryAbstract.
    """

    class Meta:
        """
        Meta class specifying the model for ExtendedEntityFactory as ExtendedEntity.
        """

        model = ExtendedEntity


class ParentEntityFactory(factories_abstract.ParentEntityFactoryAbstract):
    """
    Factory class for creating instances of ParentEntity, inheriting from
    ParentEntityFactoryAbstract.
    """

    class Meta:
        """
        Meta class specifying the model for ParentEntityFactory as ParentEntity.
        """

        model = ParentEntity



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from typing import TYPE_CHECKING

from pydantic import BaseModel

from bazis.core.utils.query_complex import QueryComplexItem


if TYPE_CHECKING:
    from .models_abstract import PermitSelectorMixin


PATTERN_SELECTORS = r'^autogen_(.+)_selectors$'
def ATTR_SELECTORS(x):  # noqa: N802
    return f'autogen_{x}_selectors'

ANONYMOUS_ID = 'anon'
PERMS_CACHE_PREFIX = 'users_perms::'
PERM_ALL = 'all'
PERM_SELF = 'self'


@enum.unique
class PermitContext(enum.Enum):
    """
    Enumeration representing different contexts in which permissions can be applied,
    such as 'item' or 'field'.
    """

    ITEM = 'item'
    FIELD = 'field'


class SelectorField(BaseModel):
    """
    Data model for a selector field, which includes the name, label, and associated
    model type.
    """

    name: str
    label: str
    model: type['PermitSelectorMixin']


class PermitStructMixin:
    """
    Mixin that provides permission handling for structures that do not inherit from
    InitialBase.
    """

    @classmethod
    def get_resource_verbose_name(cls) -> str:
        """
        Mandatory method.
        Returns a human-readable name of the structure's application, used in the permissions settings panel.
        """
        raise NotImplementedError

    @classmethod
    def get_resource_app(cls) -> str:
        """
        Mandatory method.
        Returns the programmatic name of the structure's application, used when searching for permissions.
        """
        raise NotImplementedError

    @classmethod
    def get_resource_name(cls) -> str:
        """
        Mandatory method.
        Returns the programmatic name of the structure, used when searching for permissions.
        """
        raise NotImplementedError

    @classmethod
    def get_selector_fields(cls) -> dict[str, SelectorField]:
        """
        Mandatory method.
        Defines a set of selectors applicable to the current structure. If the structure does not imply
        the presence of selectors, an empty dictionary can be returned, making only the 'all' selector applicable.
        Returns a dictionary of selector fields, where the key is the field name and the value is the selector
        model (e.g., User, Organization).
        In standard Django models, this dictionary is generated based on fields that refer to selector models.
        """
        raise NotImplementedError

    @classmethod
    def parse_selector(cls, selector: str) -> tuple[type['PermitStructMixin'], str]:
        """
        Preferable.
        Parses the selector and returns a tuple consisting of the final model and the selector name.
        Without overriding this method, selectors cannot be multi-level.
        :param kls:
        :param relation:
        :return:
        """
        return cls, selector

    @classmethod
    def selector_extending(cls, node: QueryComplexItem) -> QueryComplexItem:
        """
        Optional.
        Extends the query node for the selector.
        Can be overridden to add additional conditions for the selector.
        """
        return node

    @classmethod
    def get_default_fields(cls) -> list[str]:
        """
        Optional method.
        Returns a list of default fields of the structure that can be restricted by permissions.
        This is relevant if the structure implements field-level permissions.
        """
        return []

    def check_condition(self, conditions: dict) -> bool:
        """
        Optional method for cases where working with a specific object is not expected.
        Compares the conditions obtained from permissions with the values of the object's fields.
        Returns True if all conditions are met, otherwise False.
        Conditions: a dictionary where the key is the selector name (field name in the structure),
        and the value is the selector value, which can be a single value or a list of values.
        """
        raise NotImplementedError

    @classmethod
    def get_selector_model(cls, selector) -> type['PermitSelectorMixin']:
        """
        Does not require overriding.
        Returns the selector model (a descendant of PermitSelectorMixin) referenced by the selector.
        """
        from .models_abstract import PermitSelectorMixin

        # will return itself if the selector is 'self'
        if selector == PERM_SELF and issubclass(cls, PermitSelectorMixin):
            return cls

        kls, selector = cls.parse_selector(selector)

        if selector_field := kls.get_selector_fields().get(selector):
            return selector_field.model

    @classmethod
    def get_resource_path(cls) -> str:
        """
        Does not require overriding.
        Returns the relative path of the structure.
        """
        return f'{cls.get_resource_app()}/{cls.get_resource_name()}'

    @classmethod
    def get_resource_label(cls) -> str:
        """
        Does not require overriding.
        Returns the full label of the structure.
        """
        return f'{cls.get_resource_app()}.{cls.get_resource_name()}'



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable
from functools import reduce

from django.apps import apps
from django.contrib.gis.db import models
from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.indexes import GinIndex
from django.db.models import BooleanField, Case, QuerySet, Value, When
from django.db.models.constants import LOOKUP_SEP
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from translated_fields import TranslatedFieldWithFallback

from bazis.contrib.users.models_abstract import UserMixin
from bazis.core.models_abstract import DtMixin, InitialBase, JsonApiMixin, UuidMixin
from bazis.core.utils import triggers
from bazis.core.utils.orm import apply_calc_queryset
from bazis.core.utils.query_complex import QueryComplex, QueryComplexItem, QueryToOrm

from .schemas import ATTR_SELECTORS, PermitStructMixin, SelectorField
from .triggers import TriggerRoleCurrentInRoles, TriggerSetDefaultUserRole


class PermitModelMixin(UserMixin, PermitStructMixin):
    """
    Mixin for models that include permission handling functionality.
    """

    autogen_selectors_fields = None

    class Meta:
        """
        Meta class to define the PermitModelMixin as an abstract class.
        """

        abstract = True

    @classmethod
    def get_resource_verbose_name(cls) -> str:
        """
        Class method to retrieve the verbose name of the resource.
        """
        return cls._meta.verbose_name

    @classmethod
    def get_selector_fields(cls) -> dict[str, SelectorField]:
        """
        Class method to get selector fields for the model, filtering only those related
        to PermitSelectorMixin.
        """
        fields = {}
        for f in cls._meta.fields:
            if model := getattr(f, 'related_model', None):
                if isinstance(model, str):
                    model = apps.get_model(model)
                if issubclass(model, PermitSelectorMixin):
                    fields[f.name] = SelectorField(
                        name=f.name, label=str(f.verbose_name or f.name), model=model
                    )
        return fields

    @classmethod
    def selector_extending(cls, node: QueryComplexItem) -> QueryComplexItem:
        """
        Class method to check if the model has the specified attributes.
        """
        multi_value = node.value
        if not isinstance(multi_value, (list, set, tuple)):
            multi_value = [multi_value]

        kls, selector = cls.parse_selector(node.key)

        attr_selector = ATTR_SELECTORS(selector)
        if attr_selector in kls.get_fields_info().attributes:
            *parts, _ = node.key.split(LOOKUP_SEP)
            attr_prefix = LOOKUP_SEP.join(parts)
            if attr_prefix:
                attr_prefix = f'{attr_prefix}__'

            node |= {
                f'{attr_prefix}{attr_selector}__overlap': [
                    str(getattr(x, 'pk', x)) for x in multi_value
                ]
            }
        return node

    @classmethod
    def perms_item_apply(cls, qs: QuerySet, perms_item: Iterable) -> QuerySet:
        """
        Class method to apply permission items to a QuerySet, filtering based on
        specified conditions.
        """
        if perms_item and all(it[1] for it in perms_item):
            return QueryToOrm.qs_apply(
                qs, reduce(lambda qc, res: qc | res, [it[1] for it in perms_item])
            )
        return qs

    def check_condition(self, cond: QueryComplex) -> bool:
        """
        Method to check if the instance meets the specified conditions based on field
        values.
        """
        return getattr(self, f'_cond_{hash(cond)}', False)

    @classmethod
    def parse_selector(cls, selector) -> tuple[type['PermitModelMixin'], str]:
        *parts, selector = selector.split(LOOKUP_SEP)

        kls = cls
        for part in parts:
            rel = kls.get_fields_info().relations.get(part)
            if not rel:
                raise ValueError(f'{cls} does not have relation {part}')
            kls = rel.related_model

        return kls, selector

    def prepare_conditions(self, perms_item: Iterable):
        if perms_item:
            qs = type(self).objects.filter(pk=self.pk)
            cases = {}
            for _, cond in perms_item:
                q_orm = QueryToOrm(cond, type(self))
                if q_orm.fields_calc:
                    qs = apply_calc_queryset(qs, q_orm.fields_calc)

                if q_orm.q:
                    cases[f'_cond_{hash(cond)}'] = Case(
                        When(q_orm.q, then=Value(True)),
                        default=Value(False),
                        output_field=BooleanField()
                    )
                else:
                    cases[f'_cond_{hash(cond)}'] = Value(True)

            item = qs.annotate(**cases).first()

            for cond_name in cases.keys():
                setattr(self, cond_name, getattr(item, cond_name))

    @classmethod
    def setup_selectors_fields(cls):
        for model in PermitModelMixin.get_inheritors():
            if model._meta.abstract or model._meta.proxy:
                continue

            selectors = []
            for f_name, f in model.get_fields_info().relations.items():
                if f.reverse:
                    continue

                related_model = f.related_model
                if isinstance(related_model, str):
                    related_model = apps.get_model(related_model)

                if issubclass(related_model, PermitSelectorMixin):
                    selectors.append(f_name)

            if issubclass(model, PermitSelectorMixin):
                selectors.append('pk')

            # Filter fields by autogen_selectors_fields
            if model.autogen_selectors_fields is None:
                selectors = []
            elif model.autogen_selectors_fields:
                selectors = [f for f in selectors if f in model.autogen_selectors_fields]

            if selectors:
                model._meta.original_attrs.setdefault('indexes', [])
                if not getattr(model._meta, 'indexes', None):
                    model._meta.indexes = []

            for field_name in selectors:
                new_field = ATTR_SELECTORS(field_name)
                model.add_to_class(
                    new_field,
                    ArrayField(
                        models.CharField(max_length=255), null=True, blank=True,
                    ),
                )
                index = GinIndex(fields=[new_field])
                index.set_name_with_model(model)
                model._meta.indexes.append(index)


class PermitSelectorMixin(InitialBase):
    """
    Mixin for models that provide selector functionality for permissions.
    """

    class Meta:
        """
        Meta class to define the PermitSelectorMixin as an abstract class.
        """

        abstract = True

    @classmethod
    def get_selector_for_user(cls, user):
        """
        Class method to be implemented for retrieving selectors specific to a user.
        """
        raise NotImplementedError


SelectorField.model_rebuild()


@triggers.register(
    TriggerRoleCurrentInRoles(),
    TriggerSetDefaultUserRole(),
    with_through=True,
)
class UserPermitMixin(InitialBase):
    """
    Mixin for user models to handle roles and permissions.
    """

    roles = models.ManyToManyField(
        'permit.Role', blank=True, verbose_name=_('User roles'), related_name='users'
    )
    role_current = models.ForeignKey(
        'permit.Role',
        verbose_name=_('Current user role'),
        related_name='users_current',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )

    class Meta:
        """
        Meta class to define the UserPermitMixin as an abstract class.
        """

        abstract = True

    @classmethod
    def get_queryset(cls, **kwargs):
        """
        Class method to get the QuerySet for the user model, with related role_current
        field.
        """
        return super().get_queryset(**kwargs).select_related('role_current')

    def has_perm(self, perm, obj=None):
        """
        Method to check if the user has a specific permission. Superusers always have
        permissions.
        """
        if self.is_active and self.is_superuser:
            return True
        return False

    def has_perms(self, perm_list, obj=None):
        """
        Method to check if the user has all permissions in a list. Superusers always
        have permissions.
        """
        return all(self.has_perm(perm, obj) for perm in perm_list)

    def has_module_perms(self, app_label):
        """
        Method to check if the user has permissions for a specific app label. Superusers
        always have permissions.
        """
        if self.is_active and self.is_superuser:
            return True
        return False

    @classmethod
    def get_selector_for_user(cls, user):
        """
        Class method to retrieve the selector for a given user.
        """
        return user


class AnonymousUserPermitMixin:
    """
    Mixin for handling permissions and roles for anonymous users.
    """

    @cached_property
    def roles(self):
        """
        Cached property to retrieve roles available for anonymous users.
        """
        return apps.get_model('permit.Role').objects.filter(for_anonymous=True)

    @cached_property
    def role_current(self):
        """
        Cached property to retrieve the current role for an anonymous user.
        """
        return self.roles.first()


class BasePermissionManager(models.Manager):
    """
    Custom manager for BasePermission models, providing additional methods for
    querying.
    """

    def get_by_natural_key(self, slug):
        """
        Retrieve a BasePermission instance using its natural key (slug).
        """
        return self.get(slug=slug)


class BasePermission(JsonApiMixin, UuidMixin, DtMixin):
    """
    Abstract base model for permissions, including common fields and methods.
    """

    slug = models.CharField(_('Label'), max_length=1024, unique=True)

    class Meta:
        """
        Meta options for the BasePermission model, marking it as abstract.
        """

        abstract = True

    objects = BasePermissionManager()

    def __str__(self):
        """
        Return the string representation of the BasePermission instance, which is its
        slug.
        """
        return self.slug

    def natural_key(self) -> tuple:
        """
        Return the natural key for the BasePermission instance as a tuple containing the
        slug.
        """
        return (self.slug,)


class BaseGroup(BasePermission):
    """
    Abstract base model for groups, extending BasePermission with additional fields.
    """

    name = TranslatedFieldWithFallback(models.CharField(_('Name'), max_length=255, blank=True, default=''))

    class Meta:
        """
        Meta options for the BaseGroup model, marking it as abstract.
        """
        abstract = True

    def __str__(self):
        """
        Return the string representation of the BaseGroup instance, which is its name.
        """
        return self.name or super().__str__()


class RoleBase(BaseGroup):
    """
    Concrete model representing a role, inheriting from BaseGroup and adding
    specific fields.
    """

    groups_permission = models.ManyToManyField(
        'permit.GroupPermission',
        verbose_name=_('Permission Groups'),
        related_name='roles',
        blank=True,
    )
    for_anonymous = models.BooleanField(_('For Anonymous Users'), default=False)
    is_system = models.BooleanField(_('System Role'), default=False)

    class Meta:
        """
        Meta options for the Role model, including verbose names for the admin
        interface.
        """
        abstract = True
        verbose_name = _('Role')
        verbose_name_plural = _('Roles')


class GroupPermissionBase(BaseGroup):
    """
    Concrete model representing a group of permissions, inheriting from BaseGroup.
    """

    permissions = models.ManyToManyField(
        'Permission', verbose_name=_('Permissions'), related_name='groups', blank=True
    )

    class Meta:
        """
        Meta options for the GroupPermission model, including verbose names for the
        admin interface.
        """
        abstract = True
        verbose_name = _('Permission Group')
        verbose_name_plural = _('Permission Groups')


class PermissionBase(BasePermission):
    """
    Concrete model representing a single permission, inheriting from BasePermission.
    """

    class Meta:
        """
        Meta options for the Permission model, including verbose names for the admin
        interface.
        """
        abstract = True
        verbose_name = _('Permission')
        verbose_name_plural = _('Permissions')






# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.utils.translation import gettext_lazy as _

from bazis.core.utils.apps import BaseConfig


class PermitConfig(BaseConfig):
    """
    Configuration class for the 'permit' application within the Bazis project. It
    sets the application name and provides a human-readable name for the application
    using Django's translation utilities.
    """

    name = 'bazis.contrib.permit'
    verbose_name = _('Permit')

    def ready(self):
        """
        Method to perform actions when the application is ready, such as importing
        signals.
        """
        super().ready()

        from . import signals  # noqa: F401
        from .models_abstract import PermitModelMixin

        PermitModelMixin.setup_selectors_fields()



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import inspect
import logging
from collections import UserDict, defaultdict
from collections.abc import Callable, Iterable
from functools import reduce
from typing import Any, Self

from django.db.models import Case, IntegerField, QuerySet, Value, When
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from fastapi import Depends

from pydantic import BaseModel

from bazis.contrib.users import get_anonymous_user_model, get_user_model
from bazis.contrib.users.routes_abstract import UserRouteBase
from bazis.core.errors import JsonApi403Exception
from bazis.core.models_abstract import JsonApiMixin
from bazis.core.routes_abstract.initial import http_get, inject_make
from bazis.core.routes_abstract.jsonapi import RestrictedQsRouteMixin, with_cache_openapi_schema
from bazis.core.schemas import (
    AccessAction,
    ApiAction,
    CrudAccessAction,
    CrudApiAction,
    SchemaField,
    SchemaFields,
    SchemaResourceBuilder,
    meta_field,
)
from bazis.core.services.includes import include_to_list
from bazis.core.utils.functools import class_or_instance_method
from bazis.core.utils.orm import apply_calc_queryset
from bazis.core.utils.query_complex import QueryToOrm

from .services import PermitService
from .utils import is_model_permit


User = get_user_model()
AnonymousUser = get_anonymous_user_model()

logger = logging.getLogger(__name__)


def fields_restricts_collect(perms_field_values):
    """
    Collects and aggregates field restrictions from permission values.
    """
    fields_restricts = defaultdict(set)
    for perm_vals in perms_field_values:
        for f_name, restricts in perm_vals.items():
            for restrict in restricts.keys():
                fields_restricts[f_name] |= set(restrict.split('|'))
    return fields_restricts


class _SchemasPermitHelper:
    """
    Auxiliary class for encapsulating the schema building context.
        The schema will be assembled only if there is permission for the action.
    """

    def __init__(
        self,
        route_cls: type['PermitRouteBase'],
        permit_service: PermitService,
        api_action: ApiAction,
        item_or_model: JsonApiMixin | type[JsonApiMixin] | Callable,
        includes: list[str] = None,
        is_response_schema: bool = False,
    ):
        """
        Initializes the _SchemasPermitHelper with the necessary context for schema
        building, including route class, permit service, API action, and item or model.
        """
        item_or_model = item_or_model() if inspect.isfunction(item_or_model) else item_or_model

        self.api_action = api_action
        self.route_cls = route_cls
        self.schema_factory = route_cls.schema_factories[api_action]
        self.permit_service = permit_service
        self.permit_handler = permit_service.handler(api_action.access_action, item_or_model)
        self.item = item_or_model if isinstance(item_or_model, JsonApiMixin) else None
        self.includes = includes or []
        self.is_response_schema = is_response_schema

        if not self.permit_handler.check_access():
            raise JsonApi403Exception()

    @cached_property
    def fields(self) -> list['SchemaField']:
        """
        Based on the main instance, patches and returns a copy of the schema fields.
        """
        return self.schema_factory.fields_patch(
            fields_restricts_collect(self.permit_handler.perms_field_values)
        )

    @cached_property
    def inclusions(self) -> Iterable[type['BaseModel']]:
        """
        Performs schema assembly for existing and requested included objects.
                :return: An iterable of BaseModel types for the included objects.
        """
        fields_names = [f.name for f in self.fields]

        # we leave only those inclusions that are present in the final set of fields
        # and explicitly requested through the parameter
        inclusions_factory = {
            k: v
            for k, v in self.schema_factory.inclusions_factory_with_default.items()
            if k in fields_names and k in self.includes
        }

        for inclusion_name, inclusion_factory in inclusions_factory.items():
            # find the included object
            if not (items := self.item.fields_for_included.get(inclusion_name)):
                continue

            # process through the list, as there may be included m2m
            for item in items:
                if not item:
                    continue

                model = type(item)

                if inclusion_factory.model != model:
                    if issubclass(model, JsonApiMixin):
                        if route_cls := model.get_default_route():
                            fields_struct = route_cls.build_schema_attrs(
                                self.api_action, 'fields', SchemaFields
                            )

                            inclusion_factory = dataclasses.replace(
                                inclusion_factory,
                                route_cls=route_cls,
                                model=type(item),
                                fields_struct=fields_struct,
                            )

                fields = None

                # sub-service permissions
                if is_model_permit(type(item)):
                    permit_handler = self.permit_service.handler(
                        self.api_action.access_action, item
                    )
                    # check access rights
                    if not permit_handler.check_access():
                        continue

                    # perform field patching
                    fields = inclusion_factory.fields_patch(
                        fields_restricts_collect(permit_handler.perms_field_values)
                    )

                # assemble the included schema with a patched copy of the schema fields
                yield SchemaResourceBuilder(inclusion_factory, fields=fields).build(id=item.id)

    @cached_property
    def inclusions_for_create(self) -> Iterable[type['BaseModel']]:
        """
        Performs schema assembly for created included objects.
                :return: An iterable of BaseModel types for the created included objects.
        """
        # add schemas for creating included if they are available
        schema = self.route_cls.schema_factories[CrudApiAction.CREATE]
        inclusions_factory = {
            k: v
            for k, v in schema.inclusions_factory_with_default.items()
            if k in self.fields and k in self.includes
        }

        for _inclusion_name, inclusion_factory in inclusions_factory.items():
            # sub-service permissions
            permit_handler = self.permit_service.handler(
                CrudAccessAction.ADD, inclusion_factory.model
            )
            # check access rights
            if not permit_handler.check_access():
                continue

            # perform field patching
            fields = inclusion_factory.fields_patch(
                fields_restricts_collect(permit_handler.perms_field_values)
            )

            # assemble the included schema with a patched copy of the schema fields
            yield SchemaResourceBuilder(inclusion_factory, fields=fields).build()

    def build_schema(self, schema_resource=None, inclusions=None, is_response_schema: bool = False):
        # if the schema is not explicitly passed - it needs to be assembled with fields
        """
        Builds the schema with the given schema resource and inclusions. If no schema
        resource is explicitly passed, it assembles one with fields.
        """
        if not schema_resource:
            schema_resource = SchemaResourceBuilder(self.schema_factory, fields=self.fields).build()
        return self.schema_factory.build_schema(
            schema_resource=schema_resource, inclusions=inclusions, is_response_schema=is_response_schema
        )


class SchemasPermit(UserDict):
    """
    A dictionary-like class that manages schema permissions and building for a given
    route class, user, and item or model.
    """

    def __init__(
        self,
        route_cls: type['PermitRouteBase'],
        user: User | AnonymousUser,
        item_or_model: JsonApiMixin | type[JsonApiMixin] | Callable,
        includes: list[str] = None,
        is_response_schema: bool = False,
        dict=None,
        /,
        **kwargs,
    ):
        """
        Initializes the SchemasPermit with the route class, user, item or model, and
        optional includes.
        """
        self.route_cls = route_cls
        self.permit_service = route_cls.InjectPermit.__annotations__['permit'](user)
        self.item_or_model = item_or_model
        self.includes = includes
        self.is_response_schema = is_response_schema
        super().__init__(dict=dict, **kwargs)

    def get_helper(self, api_action: ApiAction):
        """
        Returns an instance of _SchemasPermitHelper for the given API action.
        """
        return _SchemasPermitHelper(
            self.route_cls, self.permit_service, api_action, self.item_or_model, self.includes, self.is_response_schema
        )

    def build_schema_list(self):
        """
        For the list, performs a separate assembly of the combined schema.
                Collects variations of field sets based on permissions.
                :return: The assembled schema for the list action.
        """
        helper = self.get_helper(CrudApiAction.LIST)
        schema_resources = [
            SchemaResourceBuilder(helper.schema_factory).build(is_response_schema=self.is_response_schema)]

        # get unique permission groups
        for values, cond in helper.permit_handler.perms_field_groups:
            restricts_collect = fields_restricts_collect(values)
            fields = helper.schema_factory.fields_patch(restricts_collect)

            # add a new resource schema to the set
            schema_resources.append(
                SchemaResourceBuilder(helper.schema_factory, fields=fields).build(
                    _schema_key=hash(cond), is_response_schema=self.is_response_schema
                )
            )

        if schema_resources:
            schema_resource_union = reduce(lambda it, res: it | res, schema_resources)
        else:
            schema_resource_union = None

        return helper.build_schema(inclusions=[], schema_resource=schema_resource_union)

    def build_schema_retrieve(self):
        """
        Builds the schema for the retrieve action, including any requested inclusions.
        """
        helper = self.get_helper(CrudApiAction.RETRIEVE)
        inclusions = list(helper.inclusions)
        return helper.build_schema(inclusions=inclusions, is_response_schema=self.is_response_schema)

    def build_schema_update(self):
        """
        Builds the schema for the update action, including any requested inclusions and
        those for creation.
        """
        helper = self.get_helper(CrudApiAction.UPDATE)
        return helper.build_schema(
            inclusions=list(helper.inclusions) + list(helper.inclusions_for_create)
        )

    def build_schema_create(self):
        """
        Builds the schema for the create action, including any requested inclusions for
        creation.
        """
        helper = self.get_helper(CrudApiAction.CREATE)
        return helper.build_schema(inclusions=list(helper.inclusions_for_create))

    @classmethod
    def get_builders(cls) -> dict[ApiAction, Callable]:
        """
        Returns a dictionary mapping API actions to their corresponding schema building
        methods.
        """
        return {
            CrudApiAction.LIST: cls.build_schema_list,
            CrudApiAction.RETRIEVE: cls.build_schema_retrieve,
            CrudApiAction.UPDATE: cls.build_schema_update,
            CrudApiAction.CREATE: cls.build_schema_create,
        }

    def __getitem__(self, key: ApiAction):
        """
        Retrieves the schema for the given API action, building it if it does not
        already exist in the data.
        """
        if key not in self.data:
            if get_schema := self.get_builders().get(key):
                self.data[key] = get_schema(self)
        return super().__getitem__(key)

    def get(self, key: ApiAction, default=None) -> type[BaseModel] | None:
        try:
            return self[key]
        except KeyError:
            return default


class PermitRouteBase(RestrictedQsRouteMixin, UserRouteBase):
    """
    Base class for routes that require permission checks, providing methods for
    schema management and access control.
    """

    abstract: bool = True
    schemas: SchemasPermit
    schemas_responses: SchemasPermit

    fields: dict[ApiAction, SchemaFields] = {
        CrudApiAction.CREATE: SchemaFields(
            exclude={'dt_created': None, 'dt_updated': None},
        ),
        CrudApiAction.UPDATE: SchemaFields(
            exclude={'dt_created': None, 'dt_updated': None},
        ),
    }

    @inject_make()
    class InjectPermit:
        """
        Dependency injection class for PermitService, used to manage permissions.
        """
        permit: PermitService = Depends()

    @classmethod
    def cls_init(cls):
        """
        Class-level initialization method that configures actions based on model
        properties, such as excluding creation and update actions for proxy models.
        """
        super().cls_init()
        # creation/update actions are not available for proxy models, as they can change the state
        # of visibility of proxy model objects
        if cls.model._meta.proxy:
            cls.actions_exclude = cls.actions_exclude or []
            cls.actions_exclude.extend(['action_schema_create', 'action_schema_update'])

    @http_get('/schema_list/', response_model=dict[str, Any])
    def action_schema_list(self, **kwargs):
        """
        HTTP GET endpoint to retrieve the schema for listing items, returning the
        OpenAPI schema.
        """
        return with_cache_openapi_schema(self.schemas_responses[CrudApiAction.LIST])

    @http_get('/schema_create/', response_model=dict[str, Any])
    def action_schema_create(
        self, include: str | None = Depends(include_to_list), **kwargs
    ):
        """
        HTTP GET endpoint to retrieve the schema for creating an item, returning the
        OpenAPI schema.
        """
        return with_cache_openapi_schema(self.schemas[CrudApiAction.CREATE])

    @http_get('/{item_id}/schema_retrieve/', response_model=dict[str, Any])
    def action_schema_retrieve(
        self, item_id: str, include: str | None = Depends(include_to_list), **kwargs
    ):
        """
        HTTP GET endpoint to retrieve the schema for a specific item, returning the
        OpenAPI schema.
        """
        self.set_api_action(CrudApiAction.RETRIEVE)
        self.set_item(item_id)
        return with_cache_openapi_schema(self.schemas_responses[CrudApiAction.RETRIEVE])

    @http_get('/{item_id}/schema_update/', response_model=dict[str, Any])
    def action_schema_update(
        self, item_id: str, include: str | None = Depends(include_to_list), **kwargs
    ):
        """
        HTTP GET endpoint to retrieve the schema for updating a specific item, returning
        the OpenAPI schema.
        """
        self.set_api_action(CrudApiAction.UPDATE)
        self.set_item(item_id)
        return with_cache_openapi_schema(self.schemas[CrudApiAction.UPDATE])

    def __init__(self, *args, **kwargs):
        """
        Initializes the PermitRouteBase, setting up schemas and other necessary
        properties.
        """
        super().__init__(*args, **kwargs)
        self.schemas_make()

    def schemas_make(self):
        """
        Creates and initializes the SchemasPermit instance for the route.
        """

        def get_item():
            """
            Helper function to get the current item or model for schema creation.
            """
            return self.item or self.model

        self.schemas = SchemasPermit(
            type(self), self.inject.user, get_item, getattr(self.inject, 'include', [])
        )
        self.schemas_responses = SchemasPermit(
            type(self), self.inject.user, get_item, getattr(self.inject, 'include', []), True
        )

    def check_access(
        self, access_action: AccessAction, item_or_model: JsonApiMixin | type[JsonApiMixin], item_passive=False
    ):
        """
        Checks access permissions for the given action and item or model.
                Raises a 403 exception if access is denied.
        """
        if not self.inject.permit.handler(access_action, item_or_model).check_access(
                item_passive=item_passive
        ):
            raise JsonApi403Exception()

    @class_or_instance_method
    def restrict_queryset(
        self: type[Self] | Self,
        qs: QuerySet,
        access_action: AccessAction,
        user: User = None,
        permit: PermitService = None,
    ):
        """
        Restricts the queryset based on permissions for the given access action and user.
                :return: The restricted queryset.
        """
        if not permit:
            if isinstance(self, PermitRouteBase):
                permit = self.inject.permit
            else:
                permit = [
                    f for f in dataclasses.fields(self.InjectPermit) if f.name == 'permit'
                ][0].type(user)

        permit_handler = permit.handler(access_action, qs.model)
        if not permit_handler.perms_item:
            return qs.none()

        # for the "add" action, the mere fact of having permissions is sufficient
        if access_action == CrudAccessAction.ADD:
            return qs

        # here we select by IDs of objects that are affected by field permissions
        # after obtaining such selections, we annotate the objects inside qs with hashes of specific schemas

        cases = []
        for __, cond in permit_handler.perms_field_groups:
            q_orm = QueryToOrm(cond, qs.model)
            if q_orm.fields_calc:
                qs = apply_calc_queryset(qs, q_orm.fields_calc)

            cases.append(When(q_orm.q or Value(True), then=Value(hash(cond))))

        qs = qs.annotate(
            _schema_key=Case(
                *cases,
                default=Value(None),
                output_field=IntegerField(null=True)
            )
        )

        qs = permit_handler.struct.perms_item_apply(qs, permit_handler.perms_item)
        return qs

    def get_queryset_for_list(self):
        """
        Retrieves the queryset for listing items, applying access checks and
        restrictions.
        """
        self.check_access(CrudAccessAction.VIEW, self.model)
        return self.restrict_queryset(
            super().get_queryset_for_list(),
            CrudAccessAction.VIEW,
        )

    @meta_field([CrudApiAction.LIST], title=_('Available for editing'))
    def for_change(self) -> list[str]:
        """
        Returns a list of primary keys (PKs) that will be available for editing from the
        current page.
        """
        return [
            str(pk)
            for pk in self.restrict_queryset(
                super().get_queryset_for_list().filter(pk__in=self.list()),
                CrudAccessAction.CHANGE,
            ).values_list('pk', flat=True)
        ]

    @meta_field([CrudApiAction.LIST], title=_('Available for deletion'))
    def for_delete(self) -> list[str]:
        """
        Returns a list of primary keys (PKs) that will be available for deletion from
        the current page.
        """
        return [
            str(pk)
            for pk in self.restrict_queryset(
                super().get_queryset_for_list().filter(pk__in=self.list()),
                CrudAccessAction.DELETE,
            ).values_list('pk', flat=True)
        ]

    @meta_field([CrudApiAction.LIST], title=_('Available for creation'))
    def for_create(self) -> bool:
        """
        Returns a boolean value indicating the availability of the "Create" action.
        """
        try:
            self.check_access(CrudAccessAction.ADD, self.model)
        except JsonApi403Exception:
            return False
        else:
            return True

    @meta_field([CrudApiAction.RETRIEVE], title=_('Available actions'))
    def crud_actions(self) -> list[str]:
        """
        Returns a list of available CRUD actions for the current object.
        """
        actions = []
        for action in CrudAccessAction:
            try:
                self.check_access(action, self.item)
            except JsonApi403Exception:
                pass
            else:
                actions.append(action.value)
        return actions

    def update(self, *args, **kwargs) -> JsonApiMixin:
        # in protected routes, anonymous users cannot create or update items
        """
        In protected routes, anonymous users cannot create or update items. Raises
        JsonApi403Exception if the user is anonymous.
        """
        if self.inject.user.is_anonymous:
            raise JsonApi403Exception
        return super().update(*args, **kwargs)

    def create(self, *args, **kwargs) -> JsonApiMixin:
        """
        In protected routes, anonymous users cannot create or update items. Raises
        JsonApi403Exception if the user is anonymous.
        """
        # in protected routes, anonymous users cannot create or update items
        if self.inject.user.is_anonymous:
            raise JsonApi403Exception
        return super().create(*args, **kwargs)

    def destroy(self, item_id: str) -> JsonApiMixin:
        # in protected routes, anonymous users cannot create or update items
        """
        In protected routes, anonymous users cannot create or update items. Raises
        JsonApi403Exception if the user is anonymous. Explicitly checks permissions for
        deletion.
        """
        if self.inject.user.is_anonymous:
            raise JsonApi403Exception
        item = self.set_item(item_id)
        # explicitly check permissions, as the schema is not built for deletion
        self.check_access(CrudAccessAction.DELETE, item)
        return super().destroy(item_id=item_id)

    def item_update(self, item: JsonApiMixin, data: Any):
        """
        Updates the item with the given data, checking permissions for the update action.
        """
        item = super().item_update(item, data)
        self.check_access(CrudAccessAction.CHECK, item, item_passive=True)
        return item

    def item_create(self, data: Any):
        """
        Creates an item with the given data, checking permissions for the create action.
        """
        item = super().item_create(data)
        self.check_access(CrudAccessAction.CHECK, item, item_passive=True)
        return item


# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from django import forms
from django.contrib import admin
from django.contrib.admin.views.autocomplete import AutocompleteJsonView
from django.contrib.admin.widgets import AutocompleteMixin as AutocompleteMixinDjango
from django.urls import re_path, reverse
from django.utils.html import mark_safe
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _

from translated_fields import TranslatedFieldAdmin

from bazis.contrib.permit.schemas import PATTERN_SELECTORS
from bazis.contrib.users import get_user_model
from bazis.core.admin_abstract import AutocompleteMixin, DtAdminMixin, M2mThroughMixin
from bazis.core.utils.sets_order import OrderedSet


User = get_user_model()


class UserPermitAdminMixin:
    """
    A mixin class for customizing the Django admin interface for user permits,
    adding additional fields and display options related to user roles.
    """

    def get_fields_permissions(self, request, obj=None):
        """
        Extends the parent method to include 'roles' and 'role_current' in the fields
        permissions for the admin interface.
        """
        return super().get_fields_permissions(request, obj) + ('roles', 'role_current')

    def get_list_display(self, request):
        """
        Extends the parent method to include 'get_roles_name' and 'role_current' in the
        list display for the admin interface, ensuring unique display fields using
        OrderedSet.
        """
        return tuple(
            OrderedSet(
                ('id',) + super().get_list_display(request) + ('get_roles_name', 'role_current')
            )
        )

    @admin.display(description=_('Roles'))
    def get_roles_name(self, user):
        """
        Returns a safe HTML string that displays the names of all roles associated with
        a user, separated by line breaks. This method is used in the admin interface to
        show user roles.
        """
        return mark_safe('<br/>'.join([role.name for role in user.roles.all()]))


class RoleAdminBase(M2mThroughMixin, DtAdminMixin, TranslatedFieldAdmin, admin.ModelAdmin):
    """
    Admin configuration for the Role model, including list display, editable fields,
    and horizontal filters.
    """

    list_display = (
        '__str__',
        'slug',
        'is_system',
    )
    list_editable = ('is_system',)
    filter_horizontal = ('groups_permission',)


class GroupPermissionAdminBase(DtAdminMixin, TranslatedFieldAdmin, admin.ModelAdmin):
    """
    Admin configuration for the GroupPermission model, including list display,
    editable fields, and horizontal filters.
    """

    filter_horizontal = ('permissions',)
    list_display = (
        'pk',
        'name',
        'slug',
    )
    list_editable = (
        'slug',
    )


class PermissionAdminBase(DtAdminMixin, TranslatedFieldAdmin, admin.ModelAdmin):
    """
    Admin configuration for the Permission model, including list display, editable
    fields, and search fields.
    """

    list_display = (
        '__str__',
        'slug',
    )
    search_fields = ('slug',)
    list_editable = ('slug',)


class PermitAutocompleteJsonView(AutocompleteJsonView):
    def process_request(self, request):
        if match := re.match(PATTERN_SELECTORS, request.GET['field_name']):
            target_field = match.group(1)
            request_GET = request.GET.copy() # noqa: N806
            request_GET['field_name'] = target_field
            request.GET = request_GET
        return super().process_request(request)


class PermitAutocompleteSelectMultiple(AutocompleteMixinDjango, forms.SelectMultiple):
    def __init__(self, origin_field, ref_field, admin_site, using, url_autocomplete, **kwargs):
        self.url_autocomplete = url_autocomplete
        self.origin_field = origin_field
        super().__init__(ref_field, admin_site, using=using, **kwargs)

    def build_attrs(self, base_attrs, extra_attrs=None):
        attrs = super().build_attrs(base_attrs, extra_attrs=extra_attrs)
        attrs['data-field-name'] = self.origin_field.name
        return attrs

    def get_url(self):
        return reverse(f'{self.admin_site.name}:{self.url_autocomplete}')


class PermitAutocompleteMultipleChoiceField(forms.ModelMultipleChoiceField):
    def prepare_value(self, value):
        if isinstance(value, list):
            return value
        return super().prepare_value(value)

    def clean(self, value):
        if value:
            return [str(user.pk) for user in super().clean(value)]
        return []


class PermitAutocompleteMixin(AutocompleteMixin):
    """
    Admin configuration for the ParentEntity model, including display, filter,
    search options, and related inlines.
    """

    class Media:
        js = ('admin/js/autocomplete.js',)

    def formfield_for_dbfield(self, db_field, **kwargs):
        if match := re.match(PATTERN_SELECTORS, db_field.name):
            target_field_name = match.group(1)
            origin_field = db_field
            db_field = self.model.get_fields_info().relations.get(target_field_name).model_field

            kwargs['widget'] = PermitAutocompleteSelectMultiple(
                origin_field,
                db_field,
                self.admin_site,
                using=kwargs.get("using"),
                url_autocomplete=self.get_url_autocomplete()
            )
            kwargs['label'] = capfirst(origin_field.verbose_name.strip())
            kwargs['form_class'] = PermitAutocompleteMultipleChoiceField

        formfield = super().formfield_for_dbfield(db_field, **kwargs)

        return formfield

    def get_url_autocomplete(self):
        return f'{self.opts.app_label}_{self.opts.model_name}_changelist-autocomplete'

    def get_urls(self):
        urls = super().get_urls()

        urls = [
                   re_path(
                       r'^autocomplete/$', self.autocomplete_view, name=self.get_url_autocomplete()
                       ),
               ] + urls
        return urls

    def autocomplete_view(self, request):
        return PermitAutocompleteJsonView.as_view(admin_site=self.admin_site)(request)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 5.0.1 on 2025-12-23 13:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('permit', '0005_alter_grouppermission_slug_alter_permission_slug_and_more'),
    ]

    operations = [
        migrations.RenameField(
            model_name='grouppermission',
            old_name='name',
            new_name='name_en',
        ),
        migrations.RenameField(
            model_name='role',
            old_name='name',
            new_name='name_en',
        ),
        migrations.AddField(
            model_name='grouppermission',
            name='name_ru',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Name'),
        ),
        migrations.AddField(
            model_name='role',
            name='name_ru',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Name'),
        ),
    ]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 5.0.6 on 2024-07-04 12:59

import django.contrib.auth.models
import django.contrib.auth.validators
import django.db.models.deletion
import django.utils.timezone
import model_clone.mixin
import uuid
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('permit', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('username', models.CharField(error_messages={'unique': 'A user with that username already exists.'}, help_text='Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.', max_length=150, unique=True, validators=[django.contrib.auth.validators.UnicodeUsernameValidator()], verbose_name='username')),
                ('first_name', models.CharField(blank=True, max_length=150, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=150, verbose_name='last name')),
                ('email', models.EmailField(blank=True, max_length=254, verbose_name='email address')),
                ('is_staff', models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.', verbose_name='staff status')),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.', verbose_name='active')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date joined')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.group', verbose_name='groups')),
                ('role_current', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='users_current', to='permit.role', verbose_name='Current user role')),
                ('roles', models.ManyToManyField(blank=True, related_name='users', to='permit.role', verbose_name='User roles')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.permission', verbose_name='user permissions')),
            ],
            options={
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
    ]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin

from bazis.contrib.permit.admin_abstract import UserPermitAdminMixin
from bazis.contrib.users.admin_abstract import UserAdminAbstract


@admin.register(get_user_model())
class UserAdmin(UserPermitAdminMixin, UserAdminAbstract, UserAdmin):
    pass



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pgtrigger

from bazis.core.utils.triggers import trigger_name


class TriggerRoleCurrentInRoles(pgtrigger.Trigger):
    """
    Enforce invariant: role_current must exist in roles.
    Applied to user model and its m2m through table.
    """

    name = trigger_name('role_current_in_roles')
    when = pgtrigger.Before
    operation = pgtrigger.Insert | pgtrigger.Update | pgtrigger.Delete

    def get_func(self, model):
        UserModel = self._primary_model  # noqa: N806
        roles_field = UserModel._meta.get_field('roles')
        role_current_field = UserModel._meta.get_field('role_current')

        through_model = roles_field.remote_field.through
        through_table = through_model._meta.db_table

        role_current_col = role_current_field.column
        user_pk_col = UserModel._meta.pk.column

        through_user_col = through_model._meta.get_field(UserModel._meta.model_name).column
        through_role_col = through_model._meta.get_field(
            roles_field.related_model._meta.model_name
        ).column

        if model is not UserModel:
            return """
                IF TG_OP = 'DELETE' THEN
                    RETURN OLD;
                END IF;

                RETURN NEW;
            """

        return f"""
            IF TG_OP = 'DELETE' THEN
                RETURN OLD;
            END IF;

            IF NEW.{role_current_col} IS NULL THEN
                SELECT th.{through_role_col}
                    INTO NEW.{role_current_col}
                    FROM {through_table} th
                    WHERE th.{through_user_col} = NEW.{user_pk_col}
                    ORDER BY th.{through_role_col}
                    LIMIT 1;

                IF NEW.{role_current_col} IS NULL THEN
                    RETURN NEW;
                END IF;
            END IF;

            PERFORM 1
                FROM {through_table} th
                WHERE th.{through_user_col} = NEW.{user_pk_col}
                  AND th.{through_role_col} = NEW.{role_current_col}
                LIMIT 1;

            IF NOT FOUND THEN
                -- If role_current is stale/outside roles (e.g. m2m changed before user save),
                -- normalize it to the first available role. Keep NULL if roles are empty.
                SELECT th.{through_role_col}
                    INTO NEW.{role_current_col}
                    FROM {through_table} th
                    WHERE th.{through_user_col} = NEW.{user_pk_col}
                    ORDER BY th.{through_role_col}
                    LIMIT 1;
            END IF;

            RETURN NEW;
        """


class TriggerSetDefaultUserRole(pgtrigger.Trigger):
    """
    Set role_current to the first available role when roles are added/removed.
    Applied to the user model m2m through table.
    """

    name = trigger_name('set_default_user_role')
    when = pgtrigger.After
    operation = pgtrigger.Insert | pgtrigger.Delete

    def get_func(self, model):
        UserModel = self._primary_model  # noqa: N806
        roles_field = UserModel._meta.get_field('roles')
        role_current_field = UserModel._meta.get_field('role_current')

        through_model = roles_field.remote_field.through
        through_table = through_model._meta.db_table
        user_table = UserModel._meta.db_table

        role_current_col = role_current_field.column
        user_pk_col = UserModel._meta.pk.column

        through_user_col = through_model._meta.get_field(UserModel._meta.model_name).column
        through_role_col = through_model._meta.get_field(
            roles_field.related_model._meta.model_name
        ).column

        if model is not through_model:
            return """
                IF TG_OP = 'DELETE' THEN
                    RETURN OLD;
                END IF;

                RETURN NEW;
            """

        return f"""
            IF TG_OP = 'INSERT' THEN
                UPDATE {user_table} u
                    SET {role_current_col} = s.{through_role_col}
                    FROM (
                        SELECT th.{through_role_col}
                        FROM {through_table} th
                        WHERE th.{through_user_col} = NEW.{through_user_col}
                        ORDER BY th.{through_role_col}
                        LIMIT 1
                    ) s
                    WHERE u.{user_pk_col} = NEW.{through_user_col}
                      AND u.{role_current_col} IS NULL;

                RETURN NEW;
            ELSIF TG_OP = 'DELETE' THEN
                UPDATE {user_table} u
                    SET {role_current_col} = (
                        SELECT th.{through_role_col}
                        FROM {through_table} th
                        WHERE th.{through_user_col} = OLD.{through_user_col}
                        ORDER BY th.{through_role_col}
                        LIMIT 1
                    )
                    WHERE u.{user_pk_col} = OLD.{through_user_col}
                      AND u.{role_current_col} = OLD.{through_role_col};

                RETURN OLD;
            END IF;

            RETURN NEW;
        """



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4

import pytest
from bazis_test_utils.utils import get_api_client
from translated_fields import to_attribute

from bazis.contrib.permit.models import Role
from bazis.contrib.users import get_user_model


User = get_user_model()


def create_user() -> User:
    suffix = uuid4().hex[:8]
    return User.objects.create_user(
        f'user_{suffix}',
        email=f'user_{suffix}@site.com',
        password='weak_password',
    )


def create_role(name: str) -> Role:
    return Role.objects.create(slug=name, **{to_attribute('name'): name})


def patch_user_roles(sample_app, auth_user: User, target_user: User, role_ids: list[str]):
    return get_api_client(sample_app, auth_user.jwt_build()).patch(
        f'/api/v1/users/user/{target_user.id}/',
        json_data={
            'data': {
                'id': str(target_user.id),
                'type': 'users.user',
                'bs:action': 'change',
                'attributes': {},
                'relationships': {
                    'roles': {
                        'data': [
                            {
                                'id': str(role_id),
                                'type': 'permit.role',
                            }
                            for role_id in role_ids
                        ]
                    }
                },
            },
        },
    )


@pytest.mark.django_db(transaction=True)
def test_default_role_is_set_on_first_role_add() -> None:
    user = create_user()
    role = create_role('role_first_add')

    user.roles.add(role)
    user.refresh_from_db()

    assert user.role_current_id == role.id


@pytest.mark.django_db(transaction=True)
def test_role_current_is_not_overwritten_when_already_set() -> None:
    user = create_user()
    role_1 = create_role('role_keep_current_1')
    role_2 = create_role('role_keep_current_2')

    user.roles.add(role_1)
    user.refresh_from_db()
    assert user.role_current_id == role_1.id

    user.roles.add(role_2)
    user.refresh_from_db()

    assert user.role_current_id == role_1.id


@pytest.mark.django_db(transaction=True)
def test_delete_current_role_switches_to_remaining_role() -> None:
    user = create_user()
    role_1 = create_role('role_switch_on_delete_1')
    role_2 = create_role('role_switch_on_delete_2')

    user.roles.add(role_1, role_2)
    user.role_current = role_1
    user.save(update_fields=['role_current'])

    user.roles.remove(role_1)
    user.refresh_from_db()

    assert user.role_current_id == role_2.id


@pytest.mark.django_db(transaction=True)
def test_delete_non_current_role_keeps_role_current() -> None:
    user = create_user()
    role_1 = create_role('role_keep_on_other_delete_1')
    role_2 = create_role('role_keep_on_other_delete_2')

    user.roles.add(role_1, role_2)
    user.role_current = role_2
    user.save(update_fields=['role_current'])

    user.roles.remove(role_1)
    user.refresh_from_db()

    assert user.role_current_id == role_2.id


@pytest.mark.django_db(transaction=True)
def test_delete_last_role_sets_role_current_to_null() -> None:
    user = create_user()
    role = create_role('role_last_delete')

    user.roles.add(role)
    user.refresh_from_db()
    assert user.role_current_id == role.id

    user.roles.remove(role)
    user.refresh_from_db()

    assert user.role_current_id is None


@pytest.mark.django_db(transaction=True)
def test_save_after_roles_set_normalizes_stale_role_current() -> None:
    user = create_user()
    role_1 = create_role('role_stale_save_1')
    role_2 = create_role('role_stale_save_2')

    user.roles.add(role_1, role_2)
    user.role_current = role_1
    user.save(update_fields=['role_current'])

    user.roles.set([role_2])

    # Emulate route/schemas flow: python object still has stale role_current,
    # then force_update save is executed.
    user.save(force_update=True)
    user.refresh_from_db()

    assert user.role_current_id == role_2.id


@pytest.mark.django_db(transaction=True)
def test_save_with_role_current_outside_roles_normalizes_to_first_role() -> None:
    user = create_user()
    role_valid = create_role('role_valid_for_normalization')
    role_invalid = create_role('role_invalid_for_normalization')

    user.roles.add(role_valid)
    user.role_current = role_invalid
    user.save(force_update=True)
    user.refresh_from_db()

    assert user.role_current_id == role_valid.id


@pytest.mark.django_db(transaction=True)
def test_api_patch_roles_replaces_current_with_first_from_new_roles(sample_app) -> None:
    user = create_user()
    role_1 = create_role('api_patch_switch_1')
    role_2 = create_role('api_patch_switch_2')

    user.roles.add(role_1, role_2)
    user.role_current = role_1
    user.save(update_fields=['role_current'])

    response = patch_user_roles(sample_app, user, user, [str(role_2.id)])

    assert response.status_code == 200
    user.refresh_from_db()
    assert list(user.roles.values_list('id', flat=True)) == [role_2.id]
    assert user.role_current_id == role_2.id


@pytest.mark.django_db(transaction=True)
def test_api_patch_roles_empty_sets_role_current_null(sample_app) -> None:
    user = create_user()
    role = create_role('api_patch_empty')
    user.roles.add(role)
    user.refresh_from_db()
    assert user.role_current_id == role.id

    response = patch_user_roles(sample_app, user, user, [])

    assert response.status_code == 200
    user.refresh_from_db()
    assert user.roles.count() == 0
    assert user.role_current_id is None



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from collections.abc import Iterable

from django.apps import apps
from django.conf import settings
from django.core.cache import cache
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from fastapi import Depends

from starlette.status import HTTP_403_FORBIDDEN

from bazis.contrib.users import get_anonymous_user_model, get_user_model
from bazis.contrib.users.service import get_user_optional
from bazis.core.errors import JsonApiHttpException
from bazis.core.schemas import AccessAction
from bazis.core.utils.query_complex import QueryComplex

from .schemas import (
    PERMS_CACHE_PREFIX,
    PermitContext,
    PermitStructMixin,
)
from .utils import selector_to_complex


User = get_user_model()
AnonymousUser = get_anonymous_user_model()


class PermitHandler:
    """
    Handles permission checks for a given action and item or structure, utilizing
    the PermitService.
    """

    def __init__(
        self,
        permit_service: 'PermitService',
        action: AccessAction,
        item_or_struct: PermitStructMixin | type[PermitStructMixin],
    ):
        """
        Initializes the PermitHandler with the given permit service, action, and item or
        structure.
        """
        self.permit_service = permit_service
        self.action = action

        if isinstance(item_or_struct, PermitStructMixin):
            self.struct = type(item_or_struct)
            self.item = item_or_struct
        else:
            self.struct = item_or_struct
            self.item = None

    def _get_perms(self, context: PermitContext, action: AccessAction = None):
        """
        Retrieves the permissions for a given context and action from the permit
        service.
        """
        perms = (
            self.permit_service.perms.get(self.struct.get_resource_app(), {})
            .get(self.struct.get_resource_name(), {})
            .get(context.value, {})
            .get((action or self.action).value, {})
        )
        return perms

    def _parse_perms(self, perms: dict) -> Iterable[tuple[dict, QueryComplex]]:
        """
        Parses a dictionary of permissions, searching for permissions related to the current model.
                :param perms: Dictionary of permissions.
                :return: An iterator yielding tuples of the form: permission, selector conditions.
        """
        for query_str, perm in perms.items():
            yield perm, selector_to_complex(query_str, self.permit_service.user, self.struct)

    @cached_property
    def perms_item(self) -> list[tuple[dict, QueryComplex]]:
        """
        Cached property that returns a list of parsed permissions for a specific item
        context.
        """
        perms = list(self._parse_perms(self._get_perms(PermitContext.ITEM)))
        return perms

    @cached_property
    def perms_field(self) -> list[tuple[dict, QueryComplex]]:
        """
        Cached property that returns a list of parsed permissions for a specific field
        context.
        """
        perms = list(self._parse_perms(self._get_perms(PermitContext.FIELD)))
        return perms

    def _get_perms_values(self, perms: list) -> Iterable[dict]:
        """
        Retrieves values from permissions for a specific item or in general.
                :param perms: List of permissions.
                :return: An iterable of permission values.
        """
        if self.item:
            self.item.prepare_conditions(perms)

        for perm, cond in perms:
            if not cond:
                yield perm
            elif self.item and self.item.check_condition(cond):
                yield perm

    def _get_perms_groups(self, perms: list) -> list[tuple[list[dict], QueryComplex]]:
        """
        Collects all possible groups of permission values based on selector conditions.
                :param perms: List of permissions.
                :return: A list of tuples containing grouped permission values and their conditions.
        """
        groups = defaultdict(list)
        for perm, cond in perms:
            groups[cond].append(perm)
        return [(_perms, cond) for cond, _perms in groups.items()]

    @cached_property
    def perms_item_values(self) -> list[dict]:
        """
        Cached property that returns a list of permission values for a specific item
        context.
        """
        perms = list(self._get_perms_values(self.perms_item))
        return perms

    @cached_property
    def perms_field_values(self) -> list[dict]:
        """
        Cached property that returns a list of permission values for a specific field
        context.
        """
        perms = list(self._get_perms_values(self.perms_field))
        return perms

    @cached_property
    def perms_item_groups(self) -> list[tuple[list[dict], QueryComplex]]:
        """
        Cached property that returns a list of grouped permission values for a specific
        item context.
        """
        perms = self._get_perms_groups(self.perms_item)
        return perms

    @cached_property
    def perms_field_groups(self) -> list[tuple[list[dict], QueryComplex]]:
        """
        Cached property that returns a list of grouped permission values for a specific
        field context.
        """
        perms = self._get_perms_groups(self.perms_field)
        return perms

    def check_access(self, item_passive=False) -> bool:
        """
        Checks if the action is permitted for the model based on the permissions.
                If checking for a specific entity, it verifies the presence of permission values for this item.
                If no specific entity is provided, it checks for the presence of the permissions themselves.
                item_passive: if True, then in the complete absence of permissions it is considered that everything is allowed
                :return: Boolean indicating whether access is granted.
        """
        if self.item:
            if item_passive:
                if any(self.perms_item):
                    return any(self.perms_item_values)
                else:
                    return True
            else:
                return any(self.perms_item_values)
        else:
            return any(self.perms_item)


class PermitService:
    """
    Manages permission data and provides handlers for checking permissions based on
    user roles.
    """

    handler_class = PermitHandler

    def __init__(self, user: User | AnonymousUser = Depends(get_user_optional)):
        """
        Initializes the PermitService with the given user, setting up the user and
        handlers.
        """
        self.user = user or AnonymousUser()
        self.handlers = {}

    @cached_property
    def perms(self) -> dict:
        """
        Parses string permissions into a dictionary, utilizing cache for efficiency.
                :return: A dictionary representing the permission tree.
        """
        if not hasattr(self.user, 'role_current'):
            raise JsonApiHttpException(
                HTTP_403_FORBIDDEN, detail=_('User instance anonymous or missing the role_current property')
            )

        if not self.user.role_current:
            return {}

        cache_key = f'{PERMS_CACHE_PREFIX}{self.user.role_current.slug}'

        # try to get permissions from cache
        if perms := cache.get(cache_key):
            return perms

        permissions = list(
            apps.get_model('permit.Permission')
            .objects.filter(
                groups__roles=self.user.role_current,
            )
            .distinct('id')
            .values_list('slug', flat=True)
        )

        perms = tmp = {}
        # Fill all permissions ['authing.user.item.view.all.all', ...] into a dictionary
        for it in permissions:
            for p in it.split('.'):
                tmp = tmp.setdefault(p, {})
            # indicate that the permission is complete
            tmp[None] = None
            tmp = perms

        cache.set(cache_key, perms, timeout=settings.BAZIS_PERMISSION_CACHE_EXPIRE)
        return perms

    def handler(
        self,
        access_action: AccessAction,
        item_or_struct: PermitStructMixin | type[PermitStructMixin],
    ) -> PermitHandler:
        """
        Returns a PermitHandler for the given action and item or structure, creating one
        if it doesn't exist.
        """
        handler_key = (access_action, item_or_struct)
        if handler_key not in self.handlers:
            self.handlers[handler_key] = self.handler_class(
                permit_service=self, action=access_action, item_or_struct=item_or_struct
            )
        return self.handlers[handler_key]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.db import models
from django.utils.translation import gettext_lazy as _

from bazis_test_utils.models_abstract import (
    ChildEntityBase,
    DependentEntityBase,
    ExtendedEntityBase,
    ParentEntityBase,
)

from bazis.contrib.author.models_abstract import AuthorMixin
from bazis.contrib.permit.models_abstract import PermitModelMixin
from bazis.contrib.users import get_user_model
from bazis.core.models_abstract import DtMixin, JsonApiMixin, UuidMixin
from bazis.core.triggers import FieldsTransferTrigger, FieldTransferSchema
from bazis.core.utils import triggers as bazis_triggers
from bazis.core.utils.orm import FieldJson, calc_property


User = get_user_model()


@bazis_triggers.register(
    FieldsTransferTrigger(
        related_field='parent_entities',
        fields={
            'author_parent': FieldTransferSchema(
                source='author',
            ),
        },
    )
)
class ChildEntity(
    PermitModelMixin, AuthorMixin, DtMixin, UuidMixin, JsonApiMixin, ChildEntityBase
):
    """
    Represents a child entity with a foreign key reference to a user and inherited
    mixins for permissions, authoring, timestamps, UUIDs, and JSON API integration.
    """

    author_parent = models.ForeignKey(User, blank=True, null=True, on_delete=models.SET_NULL)

    class Meta:
        """
        Meta options for the ChildEntity model including verbose names for singular and
        plural forms.
        """

        verbose_name = _('Child entity')
        verbose_name_plural = _('Child entities')


@bazis_triggers.register(
    FieldsTransferTrigger(
        related_field='parent_entity',
        fields={
            'author_parent': FieldTransferSchema(
                source='author',
            ),
        },
    )
)
class DependentEntity(
    PermitModelMixin, AuthorMixin, DtMixin, UuidMixin, JsonApiMixin, DependentEntityBase
):
    """
    Represents a dependent entity with foreign key references to a user and a parent
    entity, and inherited mixins for permissions, authoring, timestamps, UUIDs, and
    JSON API integration.
    """

    author_parent = models.ForeignKey(User, blank=True, null=True, on_delete=models.SET_NULL)
    parent_entity = models.ForeignKey(
        'ParentEntity', on_delete=models.CASCADE, related_name='dependent_entities'
    )

    class Meta:
        """
        Meta options for the DependentEntity model including verbose names for singular
        and plural forms.
        """

        verbose_name = _('Dependent entity')
        verbose_name_plural = _('Dependent entities')


@bazis_triggers.register(
    FieldsTransferTrigger(
        related_field='parent_entity',
        fields={
            'author_parent': FieldTransferSchema(
                source='author',
            ),
        },
    )
)
class ExtendedEntity(
    PermitModelMixin, AuthorMixin, DtMixin, UuidMixin, JsonApiMixin, ExtendedEntityBase
):
    """
    Represents an extended entity with a one-to-one reference to a parent entity and
    a foreign key reference to a user, along with inherited mixins for permissions,
    authoring, timestamps, UUIDs, and JSON API integration.
    """
    autogen_selectors_fields = None

    author_parent = models.ForeignKey(User, blank=True, null=True, on_delete=models.SET_NULL)
    parent_entity = models.OneToOneField(
        'ParentEntity', on_delete=models.CASCADE, related_name='extended_entity'
    )

    class Meta:
        """
        Meta options for the ExtendedEntity model including verbose names for singular
        and plural forms.
        """

        verbose_name = _('Extended entity')
        verbose_name_plural = _('Extended entities')


class ParentEntity(
    PermitModelMixin, AuthorMixin, DtMixin, UuidMixin, JsonApiMixin, ParentEntityBase
):
    """
    Represents a parent entity with a many-to-many relationship to child entities
    and inherited mixins for permissions, authoring, timestamps, UUIDs, and JSON API
    integration.
    """
    autogen_selectors_fields = ['author']

    child_entities = models.ManyToManyField(
        ChildEntity,
        related_name='parent_entities',
        blank=True,
    )

    @calc_property(
        [
            FieldJson(
                source='child_entities',
                fields=['id', 'child_name',],
            ),
        ]
    )
    def childs_detail(self) -> list:
        return [
            {
                'id': child['id'],
                'child_name': child['child_name'],
            }
            for child in self._child_entities
        ]

    @property
    def some_count_property(self) -> int:
        return 1000

    class Meta:
        """
        Meta options for the ParentEntity model including verbose names for singular and
        plural forms.
        """

        verbose_name = _('Parent entity')
        verbose_name_plural = _('Parent entities')



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import quote

import pytest
from bazis_test_utils.utils import get_api_client

from bazis.contrib.permit.models import GroupPermission, Permission, Role
from bazis.contrib.users import get_user_model

from tests import factories


User = get_user_model()

GROUPS = {
    'parent_entity': [
        'entity.parent_entity.item.add.all',  # everyone has the ability to add records
        'entity.parent_entity.item.view.author',  # only author has the ability to view his own records
        'entity.parent_entity.item.change.author',  # only the author can edit his own record
    ],
    'child_entity': [
        'entity.child_entity.item.view.all',  # everyone has the ability to view any records
        'entity.child_entity.item.change.author',  # only the author can edit the record
    ],
    'complex_parent_entity': [
        'entity.parent_entity.item.add.all',
        'entity.parent_entity.item.view.author=__selector__',
        f'entity.parent_entity.item.check.author=__selector__&name={quote("Test name 2")}',
        'entity.parent_entity.item.change.author=__selector__&is_active=false',
        'entity.parent_entity.field.view.author=__selector__&is_active=true.description.enable',
        'entity.parent_entity.field.view.author=__selector__&is_active=true.__all__.disable',
        'entity.parent_entity.field.view.author=__selector__&is_active=false.description.disable',
    ],
}


@pytest.fixture(scope='function')
def groups():
    # create roles and groups
    groups = {}
    for group_name, permissions in GROUPS.items():
        group = GroupPermission.objects.create(name_en=group_name, slug=group_name)
        for it in permissions:
            group.permissions.add(Permission.objects.get_or_create(slug=it)[0])
        groups[group_name] = group
    return groups


def role_create(groups, name: str, groups_names: list[str]):
    """
    Helper function to create a role and assign it to specified groups of
    permissions.
    """
    # create roles and groups
    role = Role.objects.create(name_en=name, slug=name)
    for group_name in groups_names:
        role.groups_permission.add(groups[group_name])
    return role


@pytest.mark.django_db(transaction=True)
def test_sparse_fieldsets(sample_app, groups):
    user_1 = User.objects.create_user('user1', email='user1@site.com', password='weak_password_2')
    parent_entity = factories.ParentEntityFactory.create(author=user_1,
                                                         description="Really hour finish thought old. Final might responsibility job result practice price possible.")
    child_entity = factories.ChildEntityFactory.create(child_name="David Bennett")
    parent_entity.child_entities.add(child_entity)
    child_entity.save()
    child_entity.refresh_from_db()

    role_1 = role_create(
        groups,
        'role_user1',
        [
            'parent_entity',
            'child_entity',
        ],
    )
    user_1.roles.add(role_1)

    # Проверка для списка.
    # Передаем поля основной модели (description), релейшенов (author), расчетных полей (childs_detail) и полей проперти (some_count_property).
    # При этом не передаем обязательный для request модели name, чтобы проверить его необязательность в response
    # Проверяем, что все остальные поля исключены.
    url = '/api/v1/entity/parent_entity/?fields[entity.parent_entity]=description,author,childs_detail,some_count_property'

    response = get_api_client(sample_app, user_1.jwt_build()).get(url)

    assert response.status_code == 200
    data = response.json()
    assert data == {
        "data": [
            {
                "attributes": {
                    "childs_detail": [
                        {
                            "child_name": "David Bennett",
                            "id": str(child_entity.id),
                        }
                    ],
                    "description": "Really hour finish thought old. Final might responsibility job result practice price possible.",
                    "some_count_property": 1000
                },
                "bs:action": "view",
                "id": str(parent_entity.id),
                "relationships": {
                    "author": {
                        "data": {
                            "id": str(user_1.id),
                            "type": "users.user"
                        }
                    }
                },
                "type": "entity.parent_entity"
            }
        ],
        "links": {
            "first": "http://testserver/api/v1/entity/parent_entity/?fields%5Bentity.parent_entity%5D=description%2Cauthor%2Cchilds_detail%2Csome_count_property",
            "last": "http://testserver/api/v1/entity/parent_entity/?fields%5Bentity.parent_entity%5D=description%2Cauthor%2Cchilds_detail%2Csome_count_property&page%5Blimit%5D=20&page%5Boffset%5D=0",
            "next": None,
            "prev": None
        },
        "meta": {}
    }

    # Проверка для объекта.
    # Передаем поля основной модели (description), релейшенов (author), расчетных полей (childs_detail) и полей проперти (some_count_property).
    # Передаем ограничение полей двух связанных моделей подключенных через include, fk и m2m связи.
    # При этом не передаем обязательный для request модели name, чтобы проверить его необязательность в response
    # Проверяем, что все остальные поля исключены.
    url = f'/api/v1/entity/parent_entity/{parent_entity.id}/?fields[entity.parent_entity]=description,author,childs_detail,some_count_property'
    url += '&include=author,child_entities&fields[users.user]=username&fields[entity.child_entity]=child_name'

    response = get_api_client(sample_app, user_1.jwt_build()).get(url)

    assert response.status_code == 200
    data = response.json()
    assert data == {
        "data": {
            "attributes": {
                "childs_detail": [
                    {
                        "child_name": "David Bennett",
                        "id": str(child_entity.id)
                    }
                ],
                "description": "Really hour finish thought old. Final might responsibility job result practice price possible.",
                "some_count_property": 1000
            },
            "bs:action": "view",
            "id": str(parent_entity.id),
            "relationships": {
                "author": {
                    "data": {
                        "id": str(user_1.id),
                        "type": "users.user"
                    }
                }
            },
            "type": "entity.parent_entity"
        },
        "included": [
            {
                "attributes": {
                    "username": "user1"
                },
                "bs:action": "view",
                "id": str(user_1.id),
                "relationships": {},
                "type": "users.user"
            },
            {
                "attributes": {
                    "child_name": "David Bennett"
                },
                "bs:action": "view",
                "id": str(child_entity.id),
                "relationships": {},
                "type": "entity.child_entity"
            }
        ],
        "meta": {}
    }

    # Проверим что удаление прав на таблицу child_entity сразу удалит связи с этой таблицей из ответа из included.
    role_1.groups_permission.remove(groups['child_entity'])

    response = get_api_client(sample_app, user_1.jwt_build()).get(url)

    assert response.status_code == 200
    data = response.json()
    assert data == {
        "data": {
            "attributes": {
                "childs_detail": [
                    {
                        "child_name": "David Bennett",
                        "id": str(child_entity.id)
                    }
                ],
                "description": "Really hour finish thought old. Final might responsibility job result practice price possible.",
                "some_count_property": 1000
            },
            "bs:action": "view",
            "id": str(parent_entity.id),
            "relationships": {
                "author": {
                    "data": {
                        "id": str(user_1.id),
                        "type": "users.user"
                    }
                }
            },
            "type": "entity.parent_entity"
        },
        "included": [
            {
                "attributes": {
                    "username": "user1"
                },
                "bs:action": "view",
                "id": str(user_1.id),
                "relationships": {},
                "type": "users.user"
            }
        ],
        "meta": {}
    }

    user_2 = User.objects.create_user('user2', email='user1@site.com', password='weak_password_2')

    # user_2 is assigned the role role_user2, which has access to the parent_entity and child_entity groups
    role_2 = role_create(
        groups,
        'role_user2',
        [
            'complex_parent_entity',
        ],
    )
    user_2.roles.add(role_2)

    # Проверка при создании по наименованию.
    # f'entity.parent_entity.item.check.author=__selector__&name={quote("Test name 2")}',
    response = get_api_client(sample_app, user_2.jwt_build()).post(
        '/api/v1/entity/parent_entity/',
        json_data={
            'data': {
                'type': 'entity.parent_entity',
                'bs:action': 'add',
                'attributes': {
                    'name': 'Test name 1',
                    'description': 'Test description',
                    'is_active': True,
                    'price': '100.49',
                    'dt_approved': '2024-01-12T16:54:12Z',
                },
            },
        },
    )
    assert response.status_code == 403
    items_id = []
    for _ in range(2):
        response = get_api_client(sample_app, user_2.jwt_build()).post(
            '/api/v1/entity/parent_entity/',
            json_data={
                'data': {
                    'type': 'entity.parent_entity',
                    'bs:action': 'add',
                    'attributes': {
                        'name': 'Test name 2',
                        'description': 'Test description',
                        'is_active': False,
                        'price': '100.49',
                        'dt_approved': '2024-01-12T16:54:12Z',
                    },
                },
            },
        )
        assert response.status_code == 201
        items_id.append(response.json()['data']['id'])

    # Проверка при обновлении по полю is_active подлежащего изменению объекта
    # 'entity.parent_entity.item.change.author=__selector__&is_active=false',
    response = get_api_client(sample_app, user_2.jwt_build()).patch(
        f'/api/v1/entity/parent_entity/{items_id[1]}/',
        json_data={
            'data': {
                'id': items_id[1],
                'type': 'entity.parent_entity',
                'bs:action': 'change',
                'attributes': {
                    'is_active': True,
                },
            },
        },
    )
    assert response.status_code == 200
    # the second attempt to change the record will be unsuccessful because only is_active=false is allowed for editing
    response = get_api_client(sample_app, user_2.jwt_build()).patch(
        f'/api/v1/entity/parent_entity/{items_id[1]}/',
        json_data={
            'data': {
                'id': items_id[1],
                'type': 'entity.parent_entity',
                'bs:action': 'change',
                'attributes': {
                    'is_active': True,
                },
            },
        },
    )
    assert response.status_code == 403

    # Проверяем что автор видит только свои записи 'entity.parent_entity.item.view.author=__selector__',
    # Проверяем что поле description включается только для is_active=true, при этом другие поля сразу отключаются.
    # И наоборот, если is_active=false, то поле description не включается, зато другие не отключаются.
    # 'entity.parent_entity.field.view.author=__selector__&is_active=true.description.enable',
    # 'entity.parent_entity.field.view.author=__selector__&is_active=true.__all__.disable',
    # 'entity.parent_entity.field.view.author=__selector__&is_active=false.description.disable',
    url = '/api/v1/entity/parent_entity/?fields[entity.parent_entity]=description,author,childs_detail,some_count_property'
    response = get_api_client(sample_app, user_2.jwt_build()).get(url)

    assert response.status_code == 200
    data = response.json()
    assert data == {
        "data": [
            {
                "attributes": {
                    "childs_detail": [],
                    "some_count_property": 1000
                },
                "bs:action": "view",
                "id": str(items_id[0]),
                "relationships": {
                    "author": {
                        "data": {
                            "id": str(user_2.id),
                            "type": "users.user"
                        }
                    }
                },
                "type": "entity.parent_entity"
            },
            {
                "attributes": {
                    "description": "Test description"
                },
                "bs:action": "view",
                "id": str(items_id[1]),
                "relationships": {},
                "type": "entity.parent_entity"
            }
        ],
        "links": {
            "first": "http://testserver/api/v1/entity/parent_entity/?fields%5Bentity.parent_entity%5D=description%2Cauthor%2Cchilds_detail%2Csome_count_property",
            "last": "http://testserver/api/v1/entity/parent_entity/?fields%5Bentity.parent_entity%5D=description%2Cauthor%2Cchilds_detail%2Csome_count_property&page%5Blimit%5D=20&page%5Boffset%5D=0",
            "next": None,
            "prev": None
        },
        "meta": {}
    }



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.apps import apps

from bazis.core.routes_abstract.jsonapi import JsonapiRouteBase


class RoleRoute(JsonapiRouteBase):
    """
    Defines the route for handling Role-related operations in the application,
    extending the JsonapiRouteBase to leverage JSON API standards. It specifies the
    model to be used and the actions that can be performed.
    """

    model = apps.get_model('permit.Role')
    actions = ['action_list', 'action_retrieve']



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.apps import apps

from bazis.contrib.author.routes_abstract import AuthorRouteBase
from bazis.contrib.permit.routes_abstract import PermitRouteBase
from bazis.core.schemas import SchemaField, SchemaFields


class ChildEntityRouteSet(PermitRouteBase, AuthorRouteBase):
    """
    Defines routing and schema fields for the ChildEntity model, including its
    relationship with parent entities.
    """

    model = apps.get_model('entity.ChildEntity')

    fields = {
        None: SchemaFields(
            include={
                'parent_entities': None,
            },
        ),
    }


class DependentEntityRouteSet(PermitRouteBase, AuthorRouteBase):
    """
    Defines routing for the DependentEntity model.
    """

    model = apps.get_model('entity.DependentEntity')


class ExtendedEntityRouteSet(PermitRouteBase, AuthorRouteBase):
    """
    Defines routing for the ExtendedEntity model.
    """

    model = apps.get_model('entity.ExtendedEntity')


class ParentEntityRouteSet(PermitRouteBase, AuthorRouteBase):
    """
    Defines routing and schema fields for the ParentEntity model, including its
    relationships with extended and dependent entities.
    """

    model = apps.get_model('entity.ParentEntity')

    # add fields (extended_entity, dependent_entities) to schema
    fields = {
        None: SchemaFields(
            include={
                'extended_entity': None, 'dependent_entities': None,
                'childs_detail': SchemaField(source='childs_detail', required=False),
                'some_count_property': SchemaField(source='some_count_property'),
            },
        ),
    }


# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import TYPE_CHECKING

from django.db.models import QuerySet

from bazis.contrib.permit.models_abstract import PermitSelectorMixin
from bazis.contrib.users import get_user_model
from bazis.core.schemas import CrudAccessAction
from bazis.core.utils.query_complex import QueryComplex, QueryComplexItem

from .schemas import PERM_ALL, PERM_SELF


if TYPE_CHECKING:
    from bazis.contrib.permit.schemas import PermitStructMixin
    from bazis.core.models_abstract import JsonApiMixin
    User = get_user_model()


def field_restrict_queryset(qs: QuerySet, context: dict) -> QuerySet:
    """
    Restrict the given queryset based on the context's route and action. If the
    context contains a route with a 'restrict_queryset' method, this method will be
    called with the queryset and action (defaulting to CrudAccessAction.VIEW).
    Otherwise, the original queryset is returned.
    """
    if 'route' in context and hasattr(context['route'], 'restrict_queryset'):
        action = context.get('restrict_action', CrudAccessAction.VIEW)
        return context['route'].restrict_queryset(qs, action)
    return qs


def is_model_permit(model: 'JsonApiMixin'):
    """
    Check if the given model's default route is a subclass of PermitRouteBase. This
    determines if the model is permitted by the application's routing rules.
    """
    from .routes_abstract import PermitRouteBase

    if issubclass(model.get_default_route(), PermitRouteBase):
        return True

    return False


def _selectors_perform(query: QueryComplex, user: 'User', struct: 'PermitStructMixin'):
    for node in (query.left, query.right):
        if isinstance(node, QueryComplexItem):
            if node.value == '__selector__':
                if node.key == PERM_ALL:
                    node.delete()

                elif not user.is_anonymous:
                    # get the selector field model
                    selector_model = struct.get_selector_model(node.key)

                    if selector_model and issubclass(selector_model, PermitSelectorMixin):
                        selector_value = selector_model.get_selector_for_user(user)

                        if selector_value:
                            # extract according to model rules
                            if node.key == PERM_SELF:
                                node.replace(key='pk', value=getattr(selector_value, 'pk', None))
                            else:
                                node.value = selector_value
                                node = struct.selector_extending(node)
                        else:
                            node.key = 'pk__isnull'
                            node.value = True

        if isinstance(node, QueryComplex):
            _selectors_perform(node, user, struct)


def selector_to_complex(query_str: str, user: 'User', struct: 'PermitStructMixin') -> QueryComplex:
    # check for simplified format
    if re.match(r'^[a-zA-Z_]+$', query_str):
        query_str = f'{query_str}=__selector__'

    query_complex = QueryComplex.from_data(query_str)
    _selectors_perform(query_complex, user, struct)

    return query_complex



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bazis.contrib.permit.models_abstract import (
    AnonymousUserPermitMixin,
    PermitSelectorMixin,
    UserPermitMixin,
)
from bazis.contrib.users.models_abstract import AnonymousUserAbstract, UserAbstract
from bazis.core.models_abstract import JsonApiMixin, UuidMixin


class User(UserPermitMixin, PermitSelectorMixin, UuidMixin, UserAbstract, JsonApiMixin):
    """
    Represents a user in the system, incorporating permissions, UUID, and user-
    specific attributes.
    """

    pass


class AnonymousUser(AnonymousUserPermitMixin, AnonymousUserAbstract):
    """
    Represents an anonymous user in the system, incorporating permissions and
    anonymous user-specific attributes.
    """

    pass



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 5.0.1 on 2024-10-24 09:04

import django.contrib.postgres.indexes
from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('entity', '0004_rename__author_parent_selectors_childentity_autogen_author_parent_selectors_and_more'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RemoveField(
            model_name='parententity',
            name='autogen_author_updated_selectors',
        ),
        migrations.AddIndex(
            model_name='childentity',
            index=django.contrib.postgres.indexes.GinIndex(fields=['autogen_author_selectors'], name='entity_chil_autogen_849c8d_gin'),
        ),
        migrations.AddIndex(
            model_name='childentity',
            index=django.contrib.postgres.indexes.GinIndex(fields=['autogen_author_updated_selectors'], name='entity_chil_autogen_c1a095_gin'),
        ),
        migrations.AddIndex(
            model_name='childentity',
            index=django.contrib.postgres.indexes.GinIndex(fields=['autogen_author_parent_selectors'], name='entity_chil_autogen_eec219_gin'),
        ),
        migrations.AddIndex(
            model_name='dependententity',
            index=django.contrib.postgres.indexes.GinIndex(fields=['autogen_author_selectors'], name='entity_depe_autogen_22be95_gin'),
        ),
        migrations.AddIndex(
            model_name='dependententity',
            index=django.contrib.postgres.indexes.GinIndex(fields=['autogen_author_updated_selectors'], name='entity_depe_autogen_c1c46a_gin'),
        ),
        migrations.AddIndex(
            model_name='dependententity',
            index=django.contrib.postgres.indexes.GinIndex(fields=['autogen_author_parent_selectors'], name='entity_depe_autogen_d06cc8_gin'),
        ),
        migrations.AddIndex(
            model_name='extendedentity',
            index=django.contrib.postgres.indexes.GinIndex(fields=['autogen_author_selectors'], name='entity_exte_autogen_55395d_gin'),
        ),
        migrations.AddIndex(
            model_name='extendedentity',
            index=django.contrib.postgres.indexes.GinIndex(fields=['autogen_author_updated_selectors'], name='entity_exte_autogen_3f6159_gin'),
        ),
        migrations.AddIndex(
            model_name='extendedentity',
            index=django.contrib.postgres.indexes.GinIndex(fields=['autogen_author_parent_selectors'], name='entity_exte_autogen_0a9cf8_gin'),
        ),
        migrations.AddIndex(
            model_name='parententity',
            index=django.contrib.postgres.indexes.GinIndex(fields=['autogen_author_selectors'], name='entity_pare_autogen_8f8317_gin'),
        ),
    ]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.apps import apps
from django.core.cache import cache
from django.db.models.signals import m2m_changed
from django.dispatch import receiver

from .schemas import PERMS_CACHE_PREFIX


@receiver(m2m_changed)
def perm_cache_clean(sender, instance, action, reverse, pk_set, *args, **kwargs):
    Role = apps.get_model('permit.Role')  # noqa: N806
    GroupPermission = apps.get_model('permit.GroupPermission') # noqa: N806
    if sender == Role.groups_permission.through:
        cache.delete(f'{PERMS_CACHE_PREFIX}{instance.slug}')
    elif sender == GroupPermission.permissions.through:
        for role in Role.objects.filter(groups_permission=instance):
            cache.delete(f'{PERMS_CACHE_PREFIX}{role.slug}')



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 5.0.6 on 2024-07-04 12:59

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('entity', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='childentity',
            name='author',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_%(class)s', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='childentity',
            name='author_parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='childentity',
            name='author_updated',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_updated_%(class)s', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='dependententity',
            name='author',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_%(class)s', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='dependententity',
            name='author_parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='dependententity',
            name='author_updated',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_updated_%(class)s', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='extendedentity',
            name='author',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_%(class)s', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='extendedentity',
            name='author_parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='extendedentity',
            name='author_updated',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_updated_%(class)s', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='parententity',
            name='author',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_%(class)s', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='parententity',
            name='author_updated',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_updated_%(class)s', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='parententity',
            name='child_entities',
            field=models.ManyToManyField(blank=True, related_name='parent_entities', to='entity.childentity'),
        ),
        migrations.AddField(
            model_name='extendedentity',
            name='parent_entity',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='extended_entity', to='entity.parententity'),
        ),
        migrations.AddField(
            model_name='dependententity',
            name='parent_entity',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='dependent_entities', to='entity.parententity'),
        ),
    ]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 5.0.1 on 2024-10-14 20:29

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('entity', '0002_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='childentity',
            name='_author_parent_selectors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, null=True, size=None),
        ),
        migrations.AddField(
            model_name='childentity',
            name='_author_selectors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, null=True, size=None),
        ),
        migrations.AddField(
            model_name='childentity',
            name='_author_updated_selectors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, null=True, size=None),
        ),
        migrations.AddField(
            model_name='dependententity',
            name='_author_parent_selectors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, null=True, size=None),
        ),
        migrations.AddField(
            model_name='dependententity',
            name='_author_selectors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, null=True, size=None),
        ),
        migrations.AddField(
            model_name='dependententity',
            name='_author_updated_selectors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, null=True, size=None),
        ),
        migrations.AddField(
            model_name='extendedentity',
            name='_author_parent_selectors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, null=True, size=None),
        ),
        migrations.AddField(
            model_name='extendedentity',
            name='_author_selectors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, null=True, size=None),
        ),
        migrations.AddField(
            model_name='extendedentity',
            name='_author_updated_selectors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, null=True, size=None),
        ),
        migrations.AddField(
            model_name='parententity',
            name='_author_selectors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, null=True, size=None),
        ),
        migrations.AddField(
            model_name='parententity',
            name='_author_updated_selectors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, null=True, size=None),
        ),
    ]
